# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import onnx
import onnx_ir as ir
from packaging.version import Version
from parameterized import parameterized

import onnxscript
import onnxscript.rewriter.testing
from onnxscript.rewriter.models import _rotary_embedding_models
from onnxscript.rewriter.rules.fusion import _rotary_embedding


class RotaryEmbeddingOnnxFusionTest(unittest.TestCase):
    @parameterized.expand(
        [
            (
                "test_case_1",
                _rotary_embedding_models.test_case_1,
            ),
            (
                "test_case_2",
                _rotary_embedding_models.test_case_2,
            ),
        ]
    )
    def test_rotary_embedding_fusion(self, _: str, test_data_constructor):
        test = test_data_constructor()
        model: ir.Model = test.get_onnx_model()
        model.graph.opset_imports[""] = 23
        model_proto = ir.serde.serialize_model(model)
        onnxscript.optimizer.optimize(model)
        _rotary_embedding.fuse_rotary_embedding(model)
        op_types = [n.op_type for n in model.graph]
        self.assertIn("RotaryEmbedding", op_types)
        rewritten_model_proto = ir.serde.serialize_model(model)
        inputs = test.get_ort_inputs()

        onnx_version = Version(onnx.__version__)
        min_version = Version("1.19.1")
        is_stable = not (onnx_version.is_devrelease or onnx_version.is_prerelease)
        if onnx_version >= min_version and is_stable:
            onnxscript.rewriter.testing.assert_numerically_equal(
                model_proto, rewritten_model_proto, args=inputs, use_reference=True
            )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import onnx.defs
import pytest

from onnxscript import ir, version_converter


class AdapterCoverageTest(unittest.TestCase):
    def get_all_unique_schema_versions(self) -> dict[str, list]:
        """Collect all unique versions of ONNX standard domain ops"""
        op_version_dict = {}
        all_schemas = onnx.defs.get_all_schemas_with_history()
        for schema in all_schemas:
            if schema.name not in op_version_dict:
                op_version_dict[schema.name] = [schema.since_version]
            else:
                if schema.since_version not in op_version_dict[schema.name]:
                    op_version_dict[schema.name].append(schema.since_version)
        return op_version_dict

    # TODO(shubhambhokare1) : Using existing onnx testing suite to verify operator adapter's functionality
    def test_upstream_coverage(self):
        op_version_dict = self.get_all_unique_schema_versions()
        op_upgrades = []
        for op_type in op_version_dict:  # pylint: disable=consider-using-dict-items
            for opset_version in op_version_dict[op_type]:
                op_upgrades.append((op_type, opset_version))

        adapter_list = version_converter._version_converter.registry.op_adapters  # pylint: disable=protected-access
        for adapter_sig in adapter_list:
            adapter_info = list(adapter_sig)
            domain, name, upgrade_version = (
                adapter_info[0],
                adapter_info[1],
                adapter_info[2] + 1,
            )
            self.assertEqual(domain, "")
            self.assertIn((name, upgrade_version), op_upgrades)

    @pytest.mark.xfail(reason="TODO: Cleanup error status API.")
    def test_version_convert_no_source_version(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "local" : 1]>
            agraph (float[4, 512, 512] input_x, float[4, 1024, 1024] input_y) => (float[4, 1024, 1024] output)
            {
                shape_a = Constant<value: tensor = int64[4] {1, 4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[4] {1, 4, 1024, 1024}>()
                reshape_y = Reshape (input_x, shape_b)
                gridsample = GridSample <mode = "bilinear"> (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[3] {4, 1024, 1024}>()
                output = Reshape (gridsample, shape_c)
            }
        """
        )
        self.assertEqual(model.graph.node(4).op_type, "GridSample")
        self.assertEqual(model.graph.node(4).attributes["mode"].value, "bilinear")

        target_version = 20
        version_converter.convert_version(model, target_version=target_version)


class VersionConverter18to17Test(unittest.TestCase):
    @pytest.mark.xfail(strict=True, reason="Version downgrade not yet supported.")
    def test_version_convert_compatible(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[1, 4, 512, 512] input_x, float[1, 4, 512, 64] input_y) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[3] {4, 512, 64}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        target_version = 17
        version_converter.convert_version(model, target_version=target_version)


class VersionConverter18to19Test(unittest.TestCase):
    def test_version_convert_compatible(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[1, 4, 512, 512] input_x, float[1, 4, 512, 64] input_y) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[3] {4, 512, 64}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        target_version = 19
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        self.assertEqual(model.graph.node(0).op_type, "Constant")
        self.assertEqual(model.graph.node(0).version, 19)
        self.assertEqual(model.graph.node(1).op_type, "Reshape")
        self.assertEqual(model.graph.node(1).version, 19)
        self.assertEqual(model.graph.node(4).op_type, "MatMul")
        self.assertEqual(model.graph.node(4).version, 19)


class VersionConverter19to20Test(unittest.TestCase):
    def test_version_convert_compatible(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[4, 512, 512] input_x) => (float[4, 257, 64, 2] output)
            {
                shape_a = Constant<value: tensor = int64[5] {1, 4, 512, 512, 1}>()
                reshape_x = Reshape (input_x, shape_a)
                dft = DFT <axis = 2, onesided = 1> (reshape_x)
                shape_c = Constant<value: tensor = int64[4] {4, 257, 64, 2}>()
                output = Reshape (dft, shape_c)
            }
        """
        )
        target_version = 20
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        self.assertEqual(model.graph.node(0).op_type, "Constant")
        self.assertEqual(model.graph.node(0).version, 20)
        self.assertEqual(model.graph.node(1).op_type, "Reshape")
        self.assertEqual(model.graph.node(1).version, 20)
        self.assertEqual(model.graph.node(2).op_type, "Constant")
        self.assertEqual(model.graph.node(3).version, 20)
        self.assertEqual(model.graph.node(3).op_type, "DFT")
        self.assertEqual(model.graph.node(3).version, 20)
        self.assertEqual(len(model.graph.node(3).inputs), 3)

    def test_version_convert_gridsample_linear(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[4, 512, 512] input_x, float[4, 1024, 1024] input_y) => (float[4, 1024, 1024] output)
            {
                shape_a = Constant<value: tensor = int64[5] {1, 4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[5] {1, 4, 1024, 1024}>()
                reshape_y = Reshape (input_x, shape_b)
                gridsample = GridSample <mode = "bilinear"> (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {4, 1024, 1024}>()
                output = Reshape (gridsample, shape_c)
            }
        """
        )
        self.assertEqual(model.graph.node(4).op_type, "GridSample")
        self.assertEqual(model.graph.node(4).attributes["mode"].value, "bilinear")

        target_version = 20
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        self.assertEqual(model.graph.node(0).op_type, "Constant")
        self.assertEqual(model.graph.node(0).version, 20)
        self.assertEqual(model.graph.node(1).op_type, "Reshape")
        self.assertEqual(model.graph.node(1).version, 20)
        self.assertEqual(model.graph.node(4).op_type, "GridSample")
        self.assertEqual(model.graph.node(4).version, 20)
        self.assertEqual(model.graph.node(4).attributes["mode"].value, "linear")

    def test_version_convert_gridsample_cubic(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[4, 512, 512] input_x, float[4, 1024, 1024] input_y) => (float[4, 1024, 1024] output)
            {
                shape_a = Constant<value: tensor = int64[5] {1, 4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[5] {1, 4, 1024, 1024}>()
                reshape_y = Reshape (input_x, shape_b)
                gridsample = GridSample <mode = "bicubic"> (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {4, 1024, 1024}>()
                output = Reshape (gridsample, shape_c)
            }
        """
        )
        self.assertEqual(model.graph.node(4).op_type, "GridSample")
        self.assertEqual(model.graph.node(4).attributes["mode"].value, "bicubic")

        target_version = 20
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        self.assertEqual(model.graph.node(0).op_type, "Constant")
        self.assertEqual(model.graph.node(0).version, 20)
        self.assertEqual(model.graph.node(1).op_type, "Reshape")
        self.assertEqual(model.graph.node(1).version, 20)
        self.assertEqual(model.graph.node(4).op_type, "GridSample")
        self.assertEqual(model.graph.node(4).version, 20)
        self.assertEqual(model.graph.node(4).attributes["mode"].value, "cubic")

    @pytest.mark.xfail(reason="Version converter does not currently support local-function.")
    def test_version_convert_function_nodes(self):
        """Test that version converter processes nodes inside model functions."""
        model = ir.from_onnx_text(
            """
            <ir_version: 8, opset_import: [ "" : 18, "pkg.custom": 1]>
            agraph (float[4, 512, 512] input_x) => (float[4, 257, 64, 2] output)
            {
                output = pkg.custom.dft_func (input_x)
            }

            <domain: "pkg.custom", opset_import: [ "" : 18]>
            dft_func (x) => (result) {
                shape_a = Constant<value: tensor = int64[5] {1, 4, 512, 512, 1}>()
                reshape_x = Reshape (x, shape_a)
                dft = DFT <axis = 2, onesided = 1> (reshape_x)
                shape_c = Constant<value: tensor = int64[4] {4, 257, 64, 2}>()
                result = Reshape (dft, shape_c)
            }
        """
        )
        # Verify the function exists with correct initial state
        self.assertEqual(len(model.functions), 1)
        func = model.functions[("pkg.custom", "dft_func", "")]
        self.assertEqual(len(func), 5)  # 5 nodes in the function

        target_version = 20
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        # Verify that the function's opset_imports are updated
        func = model.functions[("pkg.custom", "dft_func", "")]
        self.assertEqual(func.opset_imports[""], target_version)

        # Verify that nodes inside the function were version-converted
        self.assertEqual(func[0].op_type, "Constant")
        self.assertEqual(func[0].version, 20)
        self.assertEqual(func[1].op_type, "Reshape")
        self.assertEqual(func[1].version, 20)
        # After DFT adapter, a new Constant node is inserted for dft_length
        self.assertEqual(func[2].op_type, "Constant")
        self.assertEqual(func[2].version, 20)
        self.assertEqual(func[3].op_type, "DFT")
        self.assertEqual(func[3].version, 20)
        self.assertEqual(len(func[3].inputs), 3)  # DFT 19->20 adds dft_length input

    @pytest.mark.xfail(reason="Version converter does not currently support local-function.")
    def test_version_convert_function_with_control_flow_subgraph(self):
        """Test that version converter processes subgraphs inside control flow nodes in functions."""
        model = ir.from_onnx_text(
            """
            <ir_version: 8, opset_import: [ "" : 18, "pkg.custom": 1]>
            agraph (float[4, 512, 512] input_x, bool cond) => (float[4, 257, 64, 2] output)
            {
                output = pkg.custom.conditional_dft (input_x, cond)
            }

            <domain: "pkg.custom", opset_import: [ "" : 18]>
            conditional_dft (x, cond) => (result) {
                result = If (cond) <then_branch: graph = then_graph () => (out) {
                    shape_a = Constant<value: tensor = int64[5] {1, 4, 512, 512, 1}>()
                    reshape_x = Reshape (x, shape_a)
                    dft = DFT <axis = 2, onesided = 1> (reshape_x)
                    shape_c = Constant<value: tensor = int64[4] {4, 257, 64, 2}>()
                    out = Reshape (dft, shape_c)
                }, else_branch: graph = else_graph () => (out) {
                    shape_c = Constant<value: tensor = int64[4] {4, 257, 64, 2}>()
                    out = Reshape (x, shape_c)
                }>
            }
        """
        )
        # Verify the function exists with correct initial state
        self.assertEqual(len(model.functions), 1)
        func = model.functions[("pkg.custom", "conditional_dft", "")]
        self.assertEqual(len(func), 1)  # 1 node (If) in the function

        # Verify the If node has subgraphs
        if_node = func[0]
        self.assertEqual(if_node.op_type, "If")
        then_branch = if_node.attributes["then_branch"].as_graph()
        else_branch = if_node.attributes["else_branch"].as_graph()
        self.assertEqual(len(then_branch), 5)  # 5 nodes in then_branch
        self.assertEqual(len(else_branch), 2)  # 2 nodes in else_branch

        target_version = 20
        # Use internal API to test function version conversion without inlining
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        # Verify that the function's opset_imports are updated
        func = model.functions[("pkg.custom", "conditional_dft", "")]
        self.assertEqual(func.opset_imports[""], target_version)

        # Verify nodes inside the function's If node subgraphs were version-converted
        # Verify nodes inside the function's If node subgraphs were version-converted
        if_node = func[0]
        self.assertEqual(if_node.op_type, "If")
        self.assertEqual(if_node.version, 20)

        # Check then_branch subgraph nodes
        then_branch = if_node.attributes["then_branch"].as_graph()
        # After DFT adapter, a new Constant node is inserted for dft_length
        self.assertEqual(len(then_branch), 6)  # 5 + 1 new Constant for DFT
        dft_node = None
        for node in then_branch:
            self.assertEqual(node.version, 20)
            if node.op_type == "DFT":
                dft_node = node
        self.assertIsNotNone(dft_node)
        self.assertEqual(len(dft_node.inputs), 3)  # DFT 19->20 adds dft_length input

        # Check else_branch subgraph nodes
        else_branch = if_node.attributes["else_branch"].as_graph()
        self.assertEqual(len(else_branch), 2)
        for node in else_branch:
            self.assertEqual(node.version, 20)


class VersionConverter20to21Test(unittest.TestCase):
    def test_version_groupnorm(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[1, 4, 512, 512] input_x, float[2] scale, float[2] bias) => (float[4, 512, 512] output)
            {
                groupnorm = GroupNormalization <num_groups = 2> (input_x, scale, bias)
                shape_c = Constant<value: tensor = int64[4] {4, 512, 512}>()
                output = Reshape (groupnorm, shape_c)
            }
        """
        )
        target_version = 21
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        self.assertEqual(model.graph.node(3).op_type, "Reshape")
        self.assertEqual(model.graph.node(3).version, 21)
        self.assertEqual(model.graph.node(4).op_type, "Expand")
        self.assertEqual(model.graph.node(4).version, 21)
        self.assertEqual(model.graph.node(5).op_type, "Reshape")
        self.assertEqual(model.graph.node(5).version, 21)
        self.assertEqual(model.graph.node(6).op_type, "Reshape")
        self.assertEqual(model.graph.node(6).version, 21)
        self.assertEqual(model.graph.node(7).op_type, "Expand")
        self.assertEqual(model.graph.node(7).version, 21)
        self.assertEqual(model.graph.node(8).op_type, "Reshape")
        self.assertEqual(model.graph.node(8).version, 21)
        self.assertEqual(model.graph.node(9).op_type, "GroupNormalization")
        self.assertEqual(model.graph.node(9).version, 21)

    def test_version_groupnorm_no_bias(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[1, 4, 512, 512] input_x, float[2] scale) => (float[4, 512, 512] output)
            {
                groupnorm = GroupNormalization <num_groups = 2> (input_x, scale)
                shape_c = Constant<value: tensor = int64[4] {4, 512, 512}>()
                output = Reshape (groupnorm, shape_c)
            }
        """
        )
        target_version = 21
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        self.assertEqual(model.graph.node(0).op_type, "GroupNormalization")
        self.assertEqual(model.graph.node(0).version, 20)


class VersionConverterMetadataMergeTest(unittest.TestCase):
    def test_metadata_is_copied_on_version_conversion(self):
        """Test that metadata is copied from original node to replacement nodes during version conversion."""
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[4, 512, 512] input_x) => (float[4, 257, 64, 2] output)
            {
                shape_a = Constant<value: tensor = int64[5] {1, 4, 512, 512, 1}>()
                reshape_x = Reshape (input_x, shape_a)
                dft = DFT <axis = 2, onesided = 1> (reshape_x)
                shape_c = Constant<value: tensor = int64[4] {4, 257, 64, 2}>()
                output = Reshape (dft, shape_c)
            }
        """
        )
        # Find the DFT node and add metadata to it
        dft_node = model.graph.node(2)
        self.assertEqual(dft_node.op_type, "DFT")
        dft_node.metadata_props["test_key"] = "test_value"
        dft_node.metadata_props["another_key"] = "another_value"

        target_version = 25
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        # After conversion, DFT adapter adds a Constant node for axis and the DFT node is replaced
        # The replacement DFT node should have the metadata copied
        new_dft_node = model.graph.node(3)
        self.assertEqual(new_dft_node.op_type, "DFT")
        self.assertEqual(new_dft_node.version, 25)

        # Verify metadata was copied to the new DFT node
        self.assertEqual(new_dft_node.metadata_props.get("test_key"), "test_value")
        self.assertEqual(new_dft_node.metadata_props.get("another_key"), "another_value")

    def test_metadata_is_copied_to_multiple_replacement_nodes(self):
        """Test that metadata is copied to all replacement nodes when an adapter creates multiple nodes."""
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[1, 4, 512, 512] input_x, float[2] scale, float[2] bias) => (float[4, 512, 512] output)
            {
                groupnorm = GroupNormalization <num_groups = 2> (input_x, scale, bias)
                shape_c = Constant<value: tensor = int64[4] {4, 512, 512}>()
                output = Reshape (groupnorm, shape_c)
            }
        """
        )
        # Find the GroupNormalization node and add metadata to it
        groupnorm_node = model.graph.node(0)
        self.assertEqual(groupnorm_node.op_type, "GroupNormalization")
        groupnorm_node.metadata_props["source"] = "original_groupnorm"

        target_version = 21
        version_converter.convert_version(model, target_version=target_version)
        self.assertEqual(model.opset_imports[""], target_version)

        # GroupNormalization adapter creates multiple nodes (Reshape, Expand, etc.)
        # Verify that metadata was copied to the new nodes created by the adapter
        new_groupnorm_node = model.graph.node(9)
        self.assertEqual(new_groupnorm_node.op_type, "GroupNormalization")
        self.assertEqual(new_groupnorm_node.version, 21)

        # Verify metadata was copied to the new GroupNormalization node
        self.assertEqual(new_groupnorm_node.metadata_props.get("source"), "original_groupnorm")

        # Also check that intermediate nodes created by the adapter received the metadata
        # The adapter creates Reshape, Expand, Reshape nodes for scale and bias
        for i in range(9):
            node = model.graph.node(i)
            if node.version == 21 and node.op_type in ("Reshape", "Expand", "Constant"):
                self.assertEqual(
                    node.metadata_props.get("source"),
                    "original_groupnorm",
                    f"Node {i} ({node.op_type}) should have metadata copied",
                )

    @pytest.mark.xfail(reason="Version converter does not currently support local-function.")
    def test_version_convert_raises_on_function_node_with_ref_attribute(self):
        """Test that version conversion raises when a function contains a node with a ref attribute."""
        # Build a function with a LeakyRelu node that uses a RefAttr for 'alpha'
        func_input = ir.Value(name="x")
        ref_attr = ir.RefAttr("alpha", "alpha", ir.AttributeType.FLOAT)
        func_output = ir.Value(name="result")
        leaky_relu_node = ir.Node(
            domain="",
            op_type="LeakyRelu",
            inputs=[func_input],
            outputs=[func_output],
            attributes=[ref_attr],
            version=18,
        )
        func_graph = ir.Graph(
            inputs=[func_input],
            outputs=[func_output],
            nodes=[leaky_relu_node],
            opset_imports={"": 18},
        )
        func_attr_param = ir.Attr("alpha", ir.AttributeType.FLOAT, 0.01)
        function = ir.Function(
            domain="pkg.custom",
            name="leaky_relu_func",
            graph=func_graph,
            attributes=[func_attr_param],
        )

        # Build a main graph that calls the function
        main_input = ir.Value(name="input_x")
        main_output = ir.Value(name="output")
        call_node = ir.Node(
            domain="pkg.custom",
            op_type="leaky_relu_func",
            inputs=[main_input],
            outputs=[main_output],
            version=18,
        )
        main_graph = ir.Graph(
            inputs=[main_input],
            outputs=[main_output],
            nodes=[call_node],
            opset_imports={"": 18, "pkg.custom": 1},
        )
        model = ir.Model(
            main_graph,
            ir_version=8,
            functions=[function],
        )

        target_version = 20
        with self.assertRaisesRegex(
            version_converter._version_converter.VersionConverterError,  # pylint: disable=protected-access
            "has ref attribute, which is not supported by version converter",
        ):
            version_converter.convert_version(model, target_version=target_version)


class VersionConverter25to26Test(unittest.TestCase):
    @pytest.mark.xfail(strict=True, reason="Version upgrade beyond 25 not yet supported.")
    def test_version_convert_compatible(self):
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 25]>
            agraph (float[1, 4, 512, 512] input_x, float[1, 4, 512, 64] input_y) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[3] {4, 512, 64}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        target_version = 26
        version_converter.convert_version(model, target_version=target_version)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Fusion rule to remove an Expand node before a binary operator.

This implements the optimization:

    BinaryOp(Expand(x, shape), y) -> BinaryOp(x, y)
    BinaryOp(x, Expand(y, shape)) -> BinaryOp(x, y)

This is valid when the binary operator's broadcasting semantics would produce
the same output shape as first expanding the input and then applying the op.
"""

from __future__ import annotations

from onnxscript import ir
from onnxscript.rewriter._basics import MatchResult
from onnxscript.rewriter._ir_utils import get_numpy_value
from onnxscript.rewriter._rewrite_rule import RewriteRuleClassBase, RewriteRuleSet

# Binary operators in ONNX standard opset that support numpy-style broadcasting.
_BROADCAST_BINARY_OPS: tuple[str, ...] = (
    "Add",
    "And",
    "BitShift",
    "BitwiseAnd",
    "BitwiseOr",
    "BitwiseXor",
    "Div",
    "Equal",
    "Greater",
    "GreaterOrEqual",
    "Less",
    "LessOrEqual",
    "Mod",
    "Mul",
    "Or",
    "Pow",
    "PRelu",
    "Sub",
    "Xor",
)


def _compute_broadcast_dim(d1, d2):
    """Return the numpy broadcast of two dimension values.

    Each dimension value may be an ``int`` or an ``onnx_ir.SymbolicDim``.
    Returns ``None`` when the result cannot be determined statically (e.g. two
    distinct symbolic values neither of which is known to be 1).
    """
    if d1 == 1:
        return d2
    if d2 == 1:
        return d1
    if d1 == d2:
        return d1
    return None


def _compute_broadcast_shape(shape1: ir.Shape, shape2: ir.Shape) -> list | None:
    """Compute numpy-style broadcast shape symbolically.

    Returns the broadcast shape as a list of dimension values (``int`` or
    ``SymbolicDim``), or ``None`` when the result cannot be determined (e.g.
    unknown ranks or incompatible static dims).
    """
    rank1 = shape1.rank()
    rank2 = shape2.rank()
    if rank1 is None or rank2 is None:
        return None
    rank = max(rank1, rank2)
    result = []
    for i in range(rank):
        idx1 = rank1 - rank + i
        d1 = shape1[idx1] if idx1 >= 0 else 1
        idx2 = rank2 - rank + i
        d2 = shape2[idx2] if idx2 >= 0 else 1
        d = _compute_broadcast_dim(d1, d2)
        if d is None:
            return None
        result.append(d)
    return result


def _check_dims_sufficient(
    expand_shape: ir.Shape,
    x_shape: ir.Shape,
    y_shape: ir.Shape,
) -> MatchResult:
    """Check that x and y together cover every dimension of the expand target.

    For each dimension ``i`` of *expand_shape* (right-aligned) the expand is
    considered redundant when at least one of the following holds:

    - ``expand_shape[i] == 1`` - expand cannot shrink a dim, so ``x_d`` must
      also be 1 and both with and without expand produce ``y_d``.
    - ``x_d == expand_shape[i]`` - the expand is a no-op at this dim.
    - ``y_d == expand_shape[i]`` - ``y`` already supplies this expansion.

    Comparisons work for both ``int`` and ``SymbolicDim`` values.
    """
    check_result = MatchResult()
    e_rank = expand_shape.rank()
    x_rank = x_shape.rank()
    y_rank = y_shape.rank()
    if e_rank is None:
        return check_result.fail("Expand output rank is unknown.")

    for rev_i in range(e_rank):
        i = e_rank - 1 - rev_i
        e_d = expand_shape[i]

        if isinstance(e_d, int) and e_d == 1:
            continue  # expand cannot shrink; x_d is also 1, no-op

        x_idx = x_rank - 1 - rev_i
        x_d = x_shape[x_idx] if x_idx >= 0 else 1
        if x_d == e_d:
            continue  # expand is a no-op at this dimension

        y_idx = y_rank - 1 - rev_i
        y_d = y_shape[y_idx] if y_idx >= 0 else 1
        if y_d == e_d:
            continue  # y already supplies this dimension

        return check_result.fail(
            f"Cannot verify that removing Expand is safe at dimension {i}: "
            f"x_d={x_d!r}, expand_d={e_d!r}, y_d={y_d!r}."
        )

    return check_result


def _check_expand_removable(
    expand_input: ir.Value,
    shape: ir.Value,
    other_input: ir.Value,
    expand_output: ir.Value | None = None,
    binary_op_output: ir.Value | None = None,
) -> MatchResult:
    """Check if an Expand node can be safely removed before a binary op.

    The Expand ``expanded_x = Expand(x, expand_shape)`` before a binary op
    ``out = BinaryOp(expanded_x, y)`` is redundant when the binary op's own
    broadcasting produces the same output as if the expand had been applied.

    Three strategies are tried in order:

    1. **Constant expand shape** - When ``shape`` is a compile-time constant,
       the dimension values are extracted from it and the check is performed
       directly.

    2. **Expand output shape annotation** - When ``shape`` is dynamic but the
       Expand node's output value already carries a shape annotation (e.g.
       after ONNX shape inference has been applied to the model), those
       dimension values are used for the check.

    3. **Binary op output shape** - When neither of the above is available,
       the rule verifies that ``broadcast(x.shape, y.shape)`` symbolically
       equals the binary op's output shape.  If they agree, the binary op's
       own broadcasting already accounts for all the expansion and the
       Expand is redundant.

    Args:
        expand_input: The value fed into the Expand node (``x``).
        shape: The target shape operand of the Expand node.
        other_input: The other operand of the binary op (``y``).
        expand_output: The output value of the Expand node.  Required for
            strategy 2.
        binary_op_output: The output value of the binary op.  Required for
            strategy 3.

    Returns:
        A :class:`MatchResult` that is successful when the Expand can be
        removed.
    """
    check_result = MatchResult()

    x_shape = expand_input.shape
    y_shape = other_input.shape
    if x_shape is None or y_shape is None:
        return check_result.fail("Input shapes are not known.")

    x_rank = x_shape.rank()
    y_rank = y_shape.rank()

    # --- Strategy 1: expand target shape is a compile-time constant ---
    expand_shape_val = get_numpy_value(shape)
    if expand_shape_val is not None:
        expand_shape = tuple(int(v) for v in expand_shape_val.tolist())
        expand_rank = len(expand_shape)

        for rev_i in range(expand_rank):
            i = expand_rank - 1 - rev_i
            e_d = expand_shape[i]  # always a known integer from numpy

            if e_d == 1:
                continue  # expand cannot shrink; x_d is also 1, no-op

            x_idx = x_rank - 1 - rev_i
            x_d = x_shape[x_idx] if x_idx >= 0 else 1

            if isinstance(x_d, int) and x_d == e_d:
                continue  # expand is a no-op at this dimension

            y_idx = y_rank - 1 - rev_i
            y_d = y_shape[y_idx] if y_idx >= 0 else 1

            if isinstance(y_d, int) and y_d == e_d:
                continue  # y already supplies this dimension

            return check_result.fail(
                f"Cannot verify that removing Expand is safe at dimension {i}: "
                f"x_d={x_d!r}, expand_d={e_d}, y_d={y_d!r}."
            )

        return check_result

    # --- Strategy 2: Expand output shape is known (e.g. from shape inference) ---
    if expand_output is not None and expand_output.shape is not None:
        return _check_dims_sufficient(expand_output.shape, x_shape, y_shape)

    # --- Strategy 3: use the binary op's output shape ---
    # broadcast(x.shape, y.shape) must equal the binary op's output shape.
    # If it does, the binary op's own broadcasting already produces the same
    # result as first expanding x and then broadcasting.
    if binary_op_output is not None and binary_op_output.shape is not None:
        op_output_shape = binary_op_output.shape
        if op_output_shape.rank() is not None:
            computed = _compute_broadcast_shape(x_shape, y_shape)
            if computed is not None and len(computed) == op_output_shape.rank():
                if all(c == a for c, a in zip(computed, op_output_shape)):
                    return check_result
        return check_result.fail(
            "broadcast(x.shape, y.shape) does not match the binary op output shape."
        )

    return check_result.fail(
        "Expand target shape is not a constant and no shape annotations are available."
    )


class _ExpandFirstInput(RewriteRuleClassBase):
    """Removes ``BinaryOp(Expand(x, shape), y)`` -> ``BinaryOp(x, y)``."""

    def __init__(self, op_type: str) -> None:
        super().__init__(f"ExpandFirst_{op_type}", remove_nodes=False)
        self._op_type = op_type

    def pattern(self, op, x: ir.Value, shape: ir.Value, y: ir.Value) -> ir.Value:
        return getattr(op, self._op_type)(op.Expand(x, shape), y)

    def check(self, context, x: ir.Value, shape: ir.Value, y: ir.Value) -> MatchResult:
        expand_output = context.root.inputs[0] if context.root.inputs else None
        binary_op_output = context.root.outputs[0] if context.root.outputs else None
        return _check_expand_removable(
            x, shape, y, expand_output=expand_output, binary_op_output=binary_op_output
        )

    def rewrite(self, op, x: ir.Value, shape: ir.Value, y: ir.Value) -> ir.Value:
        return getattr(op, self._op_type)(x, y)


class _ExpandSecondInput(RewriteRuleClassBase):
    """Removes ``BinaryOp(x, Expand(y, shape))`` -> ``BinaryOp(x, y)``."""

    def __init__(self, op_type: str) -> None:
        super().__init__(f"ExpandSecond_{op_type}", remove_nodes=False)
        self._op_type = op_type

    def pattern(self, op, x: ir.Value, y: ir.Value, shape: ir.Value) -> ir.Value:
        return getattr(op, self._op_type)(x, op.Expand(y, shape))

    def check(self, context, x: ir.Value, y: ir.Value, shape: ir.Value) -> MatchResult:
        expand_output = context.root.inputs[1] if context.root.inputs else None
        binary_op_output = context.root.outputs[0] if context.root.outputs else None
        return _check_expand_removable(
            y, shape, x, expand_output=expand_output, binary_op_output=binary_op_output
        )

    def rewrite(self, op, x: ir.Value, y: ir.Value, shape: ir.Value) -> ir.Value:
        return getattr(op, self._op_type)(x, y)


def _make_expand_before_binary_op_rules() -> list:
    """Create rewrite rules for removing Expand before each supported binary op."""
    rules = []
    for op_type in _BROADCAST_BINARY_OPS:
        rules.append(_ExpandFirstInput.rule(op_type))
        rules.append(_ExpandSecondInput.rule(op_type))
    return rules


expand_before_binary_op_rules = RewriteRuleSet(_make_expand_before_binary_op_rules())



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import ClassVar

import onnxscript.rewriter.pattern as orp
from onnxscript import ir
from onnxscript.rewriter import _ir_utils


def _get_node(value: ir.Value, name: str) -> ir.Node:
    """Get the node from the output value."""
    node = value.producer()
    assert node is not None, f"{name} node should not be None"
    return node


def _get_kwargs(node: ir.Node) -> dict[str, float | int]:
    """Get the kwargs from the node."""
    kwargs = {key: val.value for key, val in node.attributes.items()}
    return kwargs


class FusedMatMulDiv1(orp.RewriteRuleClassBase):
    """Replaces ``MatMul + Div`` with MatMul."""

    def pattern(self, op, x, y, cst):
        return op.Div(op.MatMul(x, y), cst)

    def check(self, context, x, y, cst) -> orp.MatchResult:
        check_result = orp.MatchResult()
        if cst.const_value is None:
            return check_result.fail("Divisor is not a constant value.")
        value = cst.const_value.numpy()
        if value.size > 1:
            return check_result.fail("Divisor is not a scalar value.")
        return check_result

    def rewrite(self, op, x, y, cst):
        value = cst.const_value.numpy()
        c = float(value[0] if value.shape == (1,) else value)
        return op.FusedMatMul(x, y, alpha=1 / c, _domain="com.microsoft")


class FusedMatMulDiv2(orp.RewriteRuleClassBase):
    """Replaces ``FusedMatMul + Div`` with FusedMatMul."""

    def pattern(self, op, x, y, cst):
        return op.Div(op.FusedMatMul(x, y, _domain="com.microsoft", _outputs=["fused"]), cst)

    def check(self, context, x, y, cst, **_) -> orp.MatchResult:
        check_result = orp.MatchResult()
        if cst.const_value is None:
            return check_result.fail("Divisor is not a constant value.")
        if cst.const_value.numpy().size > 1:
            return check_result.fail("Divisor is not a scalar value.")
        return check_result

    def rewrite(self, op, x, y, cst, fused: ir.Value):
        value = cst.const_value.numpy()
        c = float(value[0] if value.shape == (1,) else value)
        fused_node = _get_node(fused, "FusedMatMul")
        kwargs = _get_kwargs(fused_node)
        kwargs["alpha"] = kwargs.get("alpha", 1.0) / c
        return op.FusedMatMul(x, y, **kwargs, _domain="com.microsoft")


class _TransposeMatMulBase(orp.RewriteRuleClassBase):
    _pos: ClassVar = 1

    def check(
        self, context, x, y, transposed: ir.Value, fused: ir.Value | None = None, **_
    ) -> orp.MatchResult:
        check_result = orp.MatchResult()
        transposed_node = _get_node(transposed, "Transpose")
        perm = transposed_node.attributes.get_ints("perm")
        if perm:
            # Check that last two dimensions are swapped
            expected_perm = list(range(len(perm)))
            expected_perm[-2], expected_perm[-1] = expected_perm[-1], expected_perm[-2]
            if list(perm) != expected_perm:
                return check_result.fail("Permutation values for Transpose are not correct.")
        elif (self._pos == 1 and not _ir_utils.has_rank(x, 2)) or (
            self._pos == 2 and not _ir_utils.has_rank(y, 2)
        ):
            # If perm is not defined, the default transpose behavior is to swap
            #   all dimensions, which is correct for MatMul with rank = 2.
            return check_result.fail(
                "If perm is not defined, rank must be 2 for TransposeMatMul rule."
            )
        if fused:
            fused_node = _get_node(fused, "FusedMatMul")
            trans_batch_property = "transBatchA" if self._pos == 1 else "transBatchB"
            if fused_node.attributes.get_int(trans_batch_property, 0):
                return check_result.fail(
                    "FusedMatMul with transposed batch cannot be used with op.Transpose in this rule."
                )
        return check_result

    def rewrite(self, op, x, y, fused: ir.Value | None = None, **_):
        kwargs = {}
        if fused:
            fused_node = _get_node(fused, "FusedMatMul")
            kwargs = _get_kwargs(fused_node)
        trans_name = "transA" if self._pos == 1 else "transB"
        kwargs[trans_name] = 1 - kwargs.get(trans_name, 0)
        return op.FusedMatMul(x, y, **kwargs, _domain="com.microsoft")


class TransposeMatMul1(_TransposeMatMulBase):
    """Replaces ``Transpose + MatMul`` with FusedMatMul."""

    def pattern(self, op, x, y):
        return op.MatMul(op.Transpose(x, _outputs=["transposed"]), y)


class TransposeFusedMatMul1(TransposeMatMul1):
    """Replaces ``Transpose + FusedMatMul`` with FusedMatMul."""

    def pattern(self, op, x, y):
        return op.FusedMatMul(
            op.Transpose(x, _outputs=["transposed"]),
            y,
            _domain="com.microsoft",
            _outputs=["fused"],
        )


class TransposeMatMul2(_TransposeMatMulBase):
    """Replaces ``Transpose + MatMul`` with FusedMatMul."""

    _pos: ClassVar = 2

    def pattern(self, op, x, y):
        return op.MatMul(x, op.Transpose(y, _outputs=["transposed"]))


class TransposeFusedMatMul2(TransposeMatMul2):
    """Replaces ``Transpose + FusedMatMul`` with FusedMatMul."""

    def pattern(self, op, x, y):
        return op.FusedMatMul(
            x,
            op.Transpose(y, _outputs=["transposed"]),
            _domain="com.microsoft",
            _outputs=["fused"],
        )


class _TransposeFusedMatMulBaseWithBatch(orp.RewriteRuleClassBase):
    """Replaces ``Transpose + FusedMatMul`` with FusedMatMul, either
       when transBatchA or transBatchB in FusedMatMul is 1, or
       can be inverted based on the permutation dims of the Transpose, in
       contrast to the original FusedMatMul rule which assumes that
       transBatchA and transBatchB are always 0 before and after rewriting.

    transBatchA = 1, transA = 0 applies a batch transpose by moving the first dimension to the second-to-last position
        i.e., equivalent to a Transpose with "perm" [1, 2, ..., N-2, 0, N-1].
    transBatchA = 0, transA = 1 flips the last two dimensions
        i.e., equivalent to a Transpose with "perm" [0, 1, ... N-3, N-1, N-2].
    transBatchA = 1, transA = 1 applies a batch transpose, then flips the last two dimensions
        i.e., equivalent to a Transpose with "perm" [1, 2, ..., N-1, 0].

    The flipping logic is based on the following cases:
       Case 1: transBatchA is 0, Transpose "perm" is [1, 2, ..., N-1, 0]
            or transBatchA is 1, Transpose "perm" is [N-1, 0, 1, ..., N-2]
        - Then transBatchA and transA can be flipped in FusedMatMul when rewriting.
       Case 2: transBatchA is 0, Transpose "perm" is [1, 2, ..., N-2, 0, N-1]
            or transBatchA is 1, Transpose "perm" is [N-2, 0, 1, ..., N-3, N-1]
        - Then transBatchA can be flipped in FusedMatMul when rewriting.
       Case 3: transBatchA is 1, Transpose "perm" is [N-1, 1, ..., N-2, 0]
        - Then transA can be flipped in FusedMatMul when rewriting.
    The same logic applies for transBatchB and transB, when _pos is set to 2.
    The _flip_transpose_batch and _flip_transpose flags are used to control
    which case is applied by the rules of inheriting classes that change these class vars.
    """

    _pos: ClassVar = 1
    _flip_transpose_batch: ClassVar = False
    _flip_transpose: ClassVar = False

    def check(
        self, context, x, y, transposed: ir.Value, fused: ir.Value, **_
    ) -> orp.MatchResult:
        check_result = orp.MatchResult()
        fused_node = _get_node(fused, "FusedMatMul")
        trans_batch_property = "transBatchA" if self._pos == 1 else "transBatchB"
        trans_batch = fused_node.attributes.get_int(trans_batch_property, 0)
        transposed_node = _get_node(transposed, "Transpose")
        perm = list(transposed_node.attributes["perm"].as_ints())
        if not perm:
            return check_result.fail("Permutation values for Transpose are not correct.")

        list_perm = list(range(len(perm)))
        if self._flip_transpose_batch and self._flip_transpose:
            #  Case 1: transBatchA/B is 0, Transpose "perm" is [1, 2, ..., N-1, 0]
            #       or transBatchA/B is 1, Transpose "perm" is [N-1, 0, 1, ..., N-2]
            #   - Then transBatchA/B and transA/B can be flipped in FusedMatMul when rewriting.
            if trans_batch == 0:
                expected_perm = [*list_perm[1:], list_perm[0]]
            else:
                expected_perm = [list_perm[-1], *list_perm[0:-1]]
            if expected_perm == perm:
                return check_result
        elif self._flip_transpose_batch:
            #  Case 2: transBatchA/B is 0, Transpose "perm" is [1, 2, ..., N-2, 0, N-1]
            #       or transBatchA/B is 1, Transpose "perm" is [N-2, 0, 1, ..., N-3, N-1]
            #   - Then transBatchA/B can be flipped in FusedMatMul when rewriting.
            if trans_batch == 0:
                expected_perm = [*list_perm[1:-1], list_perm[0], list_perm[-1]]
            else:
                expected_perm = [list_perm[-2], *list_perm[0:-2], list_perm[-1]]
            if expected_perm == perm:
                return check_result
        elif self._flip_transpose:
            #  Case 3: transBatchA is 1, Transpose "perm" is [N-1, 1, ..., N-2, 0]
            #   - Then transA can be flipped in FusedMatMul when rewriting.
            expected_perm = [list_perm[-1], *list_perm[1:-1], list_perm[0]]
            if expected_perm == perm and trans_batch == 1:
                return check_result

        return check_result.fail("Permutation values for Transpose are not correct.")

    def rewrite(self, op, x, y, fused: ir.Value, **_):
        kwargs = {}
        fused_node = _get_node(fused, "FusedMatMul")
        kwargs = _get_kwargs(fused_node)
        name = "A" if self._pos == 1 else "B"
        if self._flip_transpose_batch:
            trans_batch_property = f"transBatch{name}"
            kwargs[trans_batch_property] = 1 - kwargs.get(trans_batch_property, 0)
        if self._flip_transpose:
            trans_property = f"trans{name}"
            kwargs[trans_property] = 1 - kwargs.get(trans_property, 0)
        return op.FusedMatMul(x, y, **kwargs, _domain="com.microsoft")

    def pattern(self, op, x, y):
        if self._pos == 1:
            return op.FusedMatMul(
                op.Transpose(x, _outputs=["transposed"]),
                y,
                _domain="com.microsoft",
                _outputs=["fused"],
            )
        else:
            return op.FusedMatMul(
                x,
                op.Transpose(y, _outputs=["transposed"]),
                _domain="com.microsoft",
                _outputs=["fused"],
            )


class TransposeFusedMatMulWithFlippedBatchAndTranspose1(_TransposeFusedMatMulBaseWithBatch):
    _flip_transpose = True
    _flip_transpose_batch = True


class TransposeFusedMatMulWithFlippedBatchAndTranspose2(_TransposeFusedMatMulBaseWithBatch):
    _pos = 2
    _flip_transpose = True
    _flip_transpose_batch = True


class TransposeFusedMatMulWithFlippedBatch1(_TransposeFusedMatMulBaseWithBatch):
    _flip_transpose_batch = True


class TransposeFusedMatMulWithFlippedBatch2(_TransposeFusedMatMulBaseWithBatch):
    _pos = 2
    _flip_transpose_batch = True


class TransposeFusedMatMulWithBatchAndTranspose1(_TransposeFusedMatMulBaseWithBatch):
    _flip_transpose = True


class TransposeFusedMatMulWithBatchAndTranspose2(_TransposeFusedMatMulBaseWithBatch):
    _pos = 2
    _flip_transpose = True


class MatMulTranspose(orp.RewriteRuleClassBase):
    """Replaces ``MatMul + Transpose`` with FusedMatMul."""

    def pattern(self, op, x, y):
        return op.Transpose(op.MatMul(x, y), _outputs=["transposed"])

    def check(self, context, x, y, transposed: ir.Value, **_) -> orp.MatchResult:
        check_result = orp.MatchResult()
        transpose_node = _get_node(transposed, "Transpose")
        perm = transpose_node.attributes.get_ints("perm")
        # transA/transB only work on the last two dimensions of the input,
        # so we can only apply this rule if the inputs are rank 2.
        if _ir_utils.has_rank(x, 2) and _ir_utils.has_rank(y, 2):
            if perm:
                # Check that the two dimensions are swapped
                if tuple(perm) != (1, 0):
                    return check_result.fail(
                        "Permutation values for Transpose are not correct."
                    )
        # If perm is not defined, the default transpose behavior is to swap
        #   all dimensions, which is correct for MatMul with rank = 2.
        else:
            return check_result.fail("Rank must be 2 for MatMulTranspose rule.")
        return check_result

    def rewrite(self, op, x, y, fused: ir.Value | None = None, **_):
        kwargs = {}
        if fused:
            fused_node = _get_node(fused, "FusedMatMul")
            kwargs = _get_kwargs(fused_node)
        for name in ["transA", "transB"]:
            kwargs[name] = 1 - kwargs.get(name, 0)
        return op.FusedMatMul(y, x, **kwargs, _domain="com.microsoft")


class FusedMatMulTranspose(MatMulTranspose):
    """Replaces ``FusedMatMul + Transpose`` with FusedMatMul."""

    def pattern(self, op, x, y):
        return op.Transpose(
            op.FusedMatMul(x, y, _domain="com.microsoft", _outputs=["fused"]),
            _outputs=["transposed"],
        )


def fused_matmul_rule_sets() -> orp.RewriteRuleSet:
    """Returns a set of rules introducing onnxruntime contrib ops.
    This requires onnxruntime to run the model after it is rewritten.

    Returns:
        RewriteRuleSet
    """
    return orp.RewriteRuleSet(
        [
            FusedMatMulDiv1.rule(),
            FusedMatMulDiv2.rule(),
            FusedMatMulTranspose.rule(),
            MatMulTranspose.rule(),
            TransposeMatMul1.rule(),
            TransposeFusedMatMul1.rule(),
            TransposeMatMul2.rule(),
            TransposeFusedMatMul2.rule(),
            TransposeFusedMatMulWithFlippedBatch1.rule(),
            TransposeFusedMatMulWithFlippedBatch2.rule(),
            TransposeFusedMatMulWithFlippedBatchAndTranspose1.rule(),
            TransposeFusedMatMulWithFlippedBatchAndTranspose2.rule(),
            TransposeFusedMatMulWithBatchAndTranspose1.rule(),
            TransposeFusedMatMulWithBatchAndTranspose2.rule(),
        ]
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

__all__ = [
    "AddInitializersToInputsPass",
    "CheckerPass",
    "ClearMetadataAndDocStringPass",
    "CommonSubexpressionEliminationPass",
    "InlinePass",
    "LiftConstantsToInitializersPass",
    "LiftSubgraphInitializersToMainGraphPass",
    "RemoveInitializersFromInputsPass",
    "RemoveUnusedFunctionsPass",
    "RemoveUnusedNodesPass",
    "RemoveUnusedOpsetsPass",
    "ShapeInferencePass",
    "TopologicalSortPass",
]

from onnx_ir.passes.common import (
    AddInitializersToInputsPass,
    CheckerPass,
    ClearMetadataAndDocStringPass,
    CommonSubexpressionEliminationPass,
    InlinePass,
    LiftConstantsToInitializersPass,
    LiftSubgraphInitializersToMainGraphPass,
    RemoveInitializersFromInputsPass,
    RemoveUnusedFunctionsPass,
    RemoveUnusedNodesPass,
    RemoveUnusedOpsetsPass,
    ShapeInferencePass,
    TopologicalSortPass,
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import numpy as np
import onnx
import parameterized

from onnxscript import FLOAT, INT64, ir, opset18, script
from onnxscript.rewriter.ort_fusions import shape_optimization


def _make_model(starts: list[int], ends: list[int]) -> onnx.ModelProto:
    @script()
    def model_script(
        x: FLOAT["N"],  # noqa: F821
        dim0: INT64[1],
        dim1: INT64[1],
        dim2: INT64[1],
        dim3: INT64[1],
    ) -> INT64["M"]:  # noqa: F821
        shape = opset18.Concat(dim0, dim1, dim2, dim3, axis=0)
        reshaped = opset18.Reshape(x, shape, allowzero=1)
        transposed = opset18.Transpose(reshaped, perm=[0, 2, 1, 3])
        final_shape = opset18.Shape(transposed)
        final_dim = opset18.Slice(final_shape, starts, ends)
        return opset18.Add(final_dim, final_dim)

    model_proto = model_script.to_model_proto()
    return model_proto


# Example input data
_model_inputs = {
    "x": np.zeros((24,), dtype=np.float32),
    "dim0": np.array([2], dtype=np.int64),
    "dim1": np.array([3], dtype=np.int64),
    "dim2": np.array([4], dtype=np.int64),
    "dim3": np.array([1], dtype=np.int64),
}


class ShapeOptimizationTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            ([0], [1], "singleton"),
            ([1], [3], "two_elements"),
            ([1], [-1], "negative_index"),
            ([-2], [1000], "out_of_bounds"),
            ([-200], [-1], "negative_out_of_bounds"),
            ([2], [2], "empty_slice"),
        ]
    )
    def test_shape_optimization(self, starts: list[int], ends: list[int], _name: str):
        model_proto = _make_model(starts, ends)
        model = ir.serde.deserialize_model(model_proto)

        count = shape_optimization.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        optimized_proto = ir.serde.serialize_model(model)

        import onnxruntime as ort

        sess = ort.InferenceSession(
            model_proto.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        outputs = sess.run(None, _model_inputs)
        sess = ort.InferenceSession(
            optimized_proto.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        optimized_outputs = sess.run(None, _model_inputs)
        for orig, opt in zip(outputs, optimized_outputs):
            np.testing.assert_array_equal(orig, opt)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import numpy as np
import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

# Rewrite the computation of cos/sin cache into the form expected by ORT's custom ops.

# We match against the following code pattern:
# Original code (from transformers) for computing cos/sin cache for RoPE:
# https://github.com/huggingface/transformers/blob/0ade1caa356dce6b70ef8293addeb0898f177206/src/transformers/models/llama/modeling_llama.py#L135
# position_ids_expanded = position_ids[:, None, :].float()
# freqs = (inv_freq_expanded.float() @ position_ids_expanded.float()).transpose(1, 2)
# emb = torch.cat((freqs, freqs), dim=-1)
# cos = emb.cos()
# sin = emb.sin()
#
# We rewrite this pattern into the following form:
# inv_freq_values = inv_freq_expanded.reshape(1, -1)
# pos_id_range = np.arange(max_pos_id, dtype=np.float32).reshape(-1, 1)
# angles = np.matmul(pos_id_range, inv_freq_values)
# cos_value = np.cos(angles)
# sin_value = np.sin(angles)
# cos_2d = op.Constant(value=ir.tensor(cos_value))
# sin_2d = op.Constant(value=ir.tensor(sin_value))
#
# This produces cos/sin values in a form that can be used by ORT's custom ops.


class CosSinCacheFusion(pattern.RewriteRuleClassBase):
    def __init__(
        self,
        name: str,
        *,
        cast: bool = False,
        reshape: bool = False,
        const_freqs: bool = False,
    ):
        # This pattern makes use of shared Cos/Sin values. So, we can't remove the
        # matched nodes as part of the rewrite-step. We apply a separate final
        # pass to remove unused nodes.
        super().__init__(name, remove_nodes=False)
        # TODO: Determine what should be the default max_pos_id value
        self._max_pos_id = None
        # map from inv_freq to (cos, sin) values for transformed graph
        self._inv_freq_cos_sin_cache: dict[ir.Value, tuple[ir.Value, ir.Value]] = {}
        self._reshape = reshape
        self._cast = cast
        self._const_freqs = const_freqs

    @property
    def max_pos_id(self) -> int | None:
        return self._max_pos_id

    @max_pos_id.setter
    def max_pos_id(self, max_pos_id: int):
        self._max_pos_id = max_pos_id  # type: ignore[assignment]

    def _compute_const_freqs(self, op, angles: np.ndarray):
        """Compute cos/sin values when frequencies are constant."""
        cos_value = np.cos(angles)
        sin_value = np.sin(angles)
        cos_2d = op.Constant(value=ir.tensor(cos_value))
        sin_2d = op.Constant(value=ir.tensor(sin_value))
        return cos_2d, sin_2d

    def _compute_dynamic_freqs(self, op, inv_freq, position_ids, dtype):
        """Compute cos/sin values dynamically based on inv_freq and position_ids."""
        if self._max_pos_id is not None:
            # Use max_pos_id from the model metadata
            max_pos_id = self._max_pos_id
        elif position_ids.const_value is not None:
            # Calculate max_pos_id from the position_ids tensor
            max_pos_id = int(np.max(position_ids.const_value.numpy()))
        else:
            # Dynamically compute max_pos_id from position_ids using ONNX ops
            inv_freq = op.Reshape(inv_freq, op.Constant(value_ints=[1, -1]))
            max_pos_id = op.ReduceMax(position_ids, keepdims=0)
            max_pos_id = op.Add(max_pos_id, op.Constant(value_int=1))
            pos_id_range = op.Range(
                op.Constant(value_int=0),
                max_pos_id,
                op.Constant(value_int=1),
            )
            pos_id_range = op.Reshape(pos_id_range, op.Constant(value_ints=[-1, 1]))
            pos_id_range = op.Cast(pos_id_range, to=ir.DataType.FLOAT)
            # Compute angles and cos/sin values
            angles = op.MatMul(pos_id_range, inv_freq)
            cos_2d = op.Cos(angles)
            sin_2d = op.Sin(angles)
            return cos_2d, sin_2d

        # If we do not compute max_pos_id using ONNX ops, use inv_freq and position_ids
        # to compute angles and cos/sin values
        # Note: The one is added to max_pos_id as position_ids are 0-indexed
        # and the range of position ids should be [0, max_pos_id], max_pos_id inclusive.
        inv_freq_values = inv_freq.const_value.numpy().reshape(1, -1)
        pos_id_range = np.arange(max_pos_id + 1, dtype=np.float32).reshape(-1, 1)
        angles = np.matmul(pos_id_range, inv_freq_values)
        return self._compute_const_freqs(op, angles)

    def cleanup(self):
        self._inv_freq_cos_sin_cache.clear()

    def pattern(
        self,
        op,
        x,
        inv_freq,
        position_ids,
        interleaved,
        num_heads,
        freqs,
        dtype,
        extra_dims,
    ):
        if not self._const_freqs:
            # Compute freqs from inv_freq and position_ids. In the _const_freqs case,
            # this computation has been constant-folded away and freqs is a constant.
            # B: batch size, S: sequence length, E: embedding dimension
            # position_ids: [B, S] or [S]
            # inv_freq: [1, E, 1]
            position_ids_expanded = op.Unsqueeze(
                position_ids, extra_dims
            )  # [B, S] | [S] => [B, 1, S]
            position_ids_expanded = op.Cast(position_ids_expanded, to=ir.DataType.FLOAT)
            # if self._reshape:
            #     position_ids_expanded = op.Expand(position_ids_expanded, _allow_other_inputs=True)
            #     position_ids_expanded = op.Reshape(position_ids_expanded, _allow_other_inputs=True)
            # inv_freq may optionally be expanded to shape [B, E, 1]
            inv_freq = pattern.OrValue(
                [
                    op.Expand(inv_freq, pattern.ANY_VALUE, _outputs=["expanded_inv_freq"]),
                    inv_freq,
                ]
            )
            freqs = op.MatMul(inv_freq, position_ids_expanded)  # [B, E, S]
        # if self._reshape:
        #     freqs = op.Reshape(freqs, freqs_3d_shape)  # redundant reshape
        freqs = op.Transpose(freqs, perm=[0, 2, 1])  # [B, S, E]
        emb = op.Concat(freqs, freqs, axis=-1)
        cos = op.Cos(emb)
        if self._cast:
            cos = op.Cast(cos, to=dtype)
        sin = op.Sin(emb)
        if self._cast:
            sin = op.Cast(sin, to=dtype)
        cos_4d = op.Unsqueeze(cos, [1])  # convert
        sin_4d = op.Unsqueeze(sin, [1])
        return op.RotaryEmbedding(
            x,
            cos_4d,
            sin_4d,
            interleaved=interleaved,
            num_heads=num_heads,
            _domain="ai.onnxruntime._fusion",
        )

    def check(
        self, context, inv_freq, position_ids, freqs, extra_dims, expanded_inv_freq=None, **_
    ) -> pattern.MatchResult:  # type: ignore[name-defined]
        check_result = pattern.MatchResult()
        # TODO(rama): handle redundant reshape/expand
        if self._const_freqs:
            if (freqs.const_value is None) or not _ir_utils.has_rank(freqs, 3):
                return check_result.fail("freqs is not a constant or not 3D.", freqs)
            else:
                return check_result
        if (
            _ir_utils.has_rank(position_ids, 2) and _ir_utils.is_singleton_value(extra_dims, 1)
        ) or (
            _ir_utils.has_rank(position_ids, 1) and _ir_utils.is_1d_value(extra_dims, [0, 1])
        ):
            pass
        else:
            return check_result.fail("position_ids is not a 1D or 2D tensor.", position_ids)
        if not _ir_utils.has_rank(inv_freq, 3):
            return check_result.fail("inv_freq is not 3D.", inv_freq)
        inv_freq_shape = inv_freq.shape
        if expanded_inv_freq is not None:
            if not _ir_utils.has_rank(expanded_inv_freq, 3):
                return check_result.fail("expanded_inv_freq is not 3D.", expanded_inv_freq)
            # TODO: check expanded_inv_freq shape
        if inv_freq.const_value is None:  # TODO: should this be inv_freq_shape?
            return check_result.fail("inv_freq is not a constant.", inv_freq)
        if inv_freq_shape[0] != 1 or inv_freq_shape[2] != 1:
            return check_result.fail("inv_freq is not of shape [1, ., 1].", inv_freq)
        return check_result

    def rewrite(
        self, op, x, inv_freq, position_ids, interleaved, num_heads, freqs, dtype, **_
    ):
        if inv_freq in self._inv_freq_cos_sin_cache:
            cos_2d, sin_2d = self._inv_freq_cos_sin_cache[inv_freq]
        else:
            # Compute cos/sin values based on whether frequencies are constant
            if self._const_freqs:
                cos_2d, sin_2d = self._compute_const_freqs(op, freqs.const_value.numpy())
            else:
                cos_2d, sin_2d = self._compute_dynamic_freqs(op, inv_freq, position_ids, dtype)
            if self._cast:
                cos_2d = op.Cast(cos_2d, to=dtype)
                sin_2d = op.Cast(sin_2d, to=dtype)
            self._inv_freq_cos_sin_cache[inv_freq] = (cos_2d, sin_2d)
        if _ir_utils.has_rank(position_ids, 1):
            zero_1d = op.Constant(value_ints=[0])
            position_ids = op.Unsqueeze(position_ids, zero_1d)
        return op.RotaryEmbedding(
            x,
            position_ids,
            cos_2d,
            sin_2d,
            interleaved=interleaved,
            num_heads=num_heads,
            _domain="com.microsoft",
        )


_cast_const_freqs = CosSinCacheFusion.rule(
    "CosSinCache_cast_const_freqs", cast=True, const_freqs=True
)
_cast = CosSinCacheFusion.rule("CosSinCache_cast", cast=True, const_freqs=False)
_const_freqs = CosSinCacheFusion.rule("CosSinCache_const_freqs", cast=False, const_freqs=True)
_basic = CosSinCacheFusion.rule("CosSinCache", cast=False)

cos_sin_cache_rules = pattern.RewriteRuleSet([_cast, _cast_const_freqs, _const_freqs, _basic])


fuse_cos_sin_cache = _fusion_utils.apply_fusion_rules(cos_sin_cache_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import onnx_ir.passes.common as common_passes
import packaging.version

import onnxscript.optimizer
import onnxscript.rewriter.ort_fusions._core as xformers
from onnxscript.rewriter.models._phi2lm import phi2lm_test
from onnxscript.rewriter.models._smollm_2 import smollm_test_2
from onnxscript.rewriter.models._whisper_decoder import whisper_decoder_test
from onnxscript.rewriter.models._whisper_encoder import whisper_encoder_test
from onnxscript.rewriter.ort_fusions._test_utils import ORT_VERSION, assert_allclose, ort_run


class TestMultiHeadAttention(unittest.TestCase):
    def test_smollm(self):
        # Generate model
        smollm_test = smollm_test_2()
        model = smollm_test.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        xformers.fuse_rms_normalization(model)
        xformers.fuse_skip_rms_normalization(model)
        xformers.fuse_rotary_embedding(model)
        xformers.fuse_cos_sin_cache(model)

        test_with_ort = packaging.version.Version("1.20") <= ORT_VERSION
        if test_with_ort:
            # Run model
            inputs = smollm_test.get_ort_inputs()
            original_outputs = ort_run("original", model, inputs)

        # Fuse SDPA and MHA
        sdpa_count = xformers.fuse_sdpa(model)
        self.assertGreater(sdpa_count, 0)
        mha_count = xformers.fuse_mha1(model)
        mha_count += xformers.fuse_mha2(model)
        self.assertGreater(mha_count, 0)

        if test_with_ort:
            # Run model again
            new_outputs = ort_run("optimized", model, inputs)
            assert_allclose(new_outputs, original_outputs)

    def test_whisper_encoder(self):
        # Generate model
        whisper_encoder = whisper_encoder_test()
        model = whisper_encoder.get_onnx_model()
        onnxscript.optimizer.optimize(model)

        test_with_ort = packaging.version.Version("1.20") <= ORT_VERSION
        if test_with_ort:
            # Run model
            inputs = whisper_encoder.get_ort_inputs()
            original_outputs = ort_run("original", model, inputs)

        # Fuse SDPA and MHA
        sdpa_count = xformers.fuse_sdpa(model, debug=True)
        self.assertGreater(sdpa_count, 0)
        model = common_passes.ShapeInferencePass()(model).model
        mha_count = xformers.fuse_mha1(model)
        mha_count += xformers.fuse_mha2(model)
        self.assertGreater(mha_count, 0)
        onnxscript.optimizer.optimize(model)

        if test_with_ort:
            # Run model again
            new_outputs = ort_run("optimized", model, inputs)
            assert_allclose(new_outputs, original_outputs)

    def test_whisper_decoder(self):
        # Generate model
        whisper_decoder = whisper_decoder_test()
        model = whisper_decoder.get_onnx_model()
        onnxscript.optimizer.optimize(model)

        test_with_ort = packaging.version.Version("1.20") <= ORT_VERSION
        if test_with_ort:
            # Run model
            inputs = whisper_decoder.get_ort_inputs()
            original_outputs = ort_run("original", model, inputs)

        # Fuse SDPA and MHA
        sdpa_count = xformers.fuse_sdpa(model)
        self.assertGreater(sdpa_count, 0)
        model = common_passes.ShapeInferencePass()(model).model
        mha_count = xformers.fuse_mha1(model)
        mha_count += xformers.fuse_mha2(model)
        self.assertGreater(mha_count, 0)
        onnxscript.optimizer.optimize(model)

        if test_with_ort:
            # Run model again
            new_outputs = ort_run("optimized", model, inputs)
            assert_allclose(new_outputs, original_outputs)

    def test_phi2lm(self):
        test_case = phi2lm_test()
        model = test_case.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        xformers.optimize_for_ort(model)
        mha_nodes = [n for n in model.graph if n.op_type == "MultiHeadAttention"]
        self.assertEqual(
            len(mha_nodes),
            1,
            "Expected exactly one MultiHeadAttention node after optimization",
        )
        mha_node = mha_nodes[0]
        # Check that the MHA node has past kv cache inputs
        self.assertEqual(len(mha_node.inputs), 8, "Expected MHA node to have 8 inputs")
        self.assertIsNotNone(mha_node.inputs[6], "Expected MHA node to have past key input")
        self.assertIsNotNone(mha_node.inputs[7], "Expected MHA node to have past value input")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

# Add first version of the RotaryEmbeddingFusion rule. This considers only one simple pattern
# for full rotation without interleaving.
# TODO(rama): Add pattern variations to handle other cases (interleaved, as well as partial rotation).

# Note: This targets the new op being proposed to ONNX. This version does not exist in ORT yet.
# so it can't be tested by running against ORT. See cos_sin_cache.py for a transformation that
# rewrites the pattern into one that can be run against ORT.


def _rotate_half_pattern(op, x, start1, end1, start2, end2):
    # Slice(input, starts, ends, axes, steps)
    x1 = op.Slice(x, start1, end1, [3], [1])
    x2 = op.Slice(x, start2, end2, [3], [1])
    minus_x2 = op.Neg(x2)
    rotated_x = op.Concat(minus_x2, x1, axis=-1)
    return rotated_x


class RotaryEmbeddingFusion(pattern.RewriteRuleClassBase):
    def __init__(self):
        super().__init__(name="RotaryEmbedding", as_function=True)

    def pattern(self, op, x, cos, sin, start1, end1, start2, end2):
        return x * cos + _rotate_half_pattern(op, x, start1, end1, start2, end2) * sin

    def check(self, op, x, start1, end1, start2, end2, **_) -> pattern.MatchResult:  # type: ignore[name-defined]
        check_result = pattern.MatchResult()
        # x needs to be a 4D tensor with known last dimension size (== head_size) and known second dimension (num_heads)
        if x is None or x.shape is None or len(x.shape) != 4:
            return check_result.fail("Input is not a 4D tensor.", x)
        if not isinstance(x.shape[1], int):
            return check_result.fail("Input dimension 1 is not an integer.", x)
        head_size = x.shape[3]
        if not isinstance(head_size, int):
            return check_result.fail("Head size is not an integer.", x)
        half_head_size = head_size // 2

        # Check that x is being split into two equal halves of size half_head_size
        if not (
            _ir_utils.is_singleton_value(start1, 0)
            and _ir_utils.is_singleton_value(end1, half_head_size)
            and _ir_utils.is_singleton_value(start2, half_head_size)
            and _ir_utils.is_singleton_value(end2, lambda x: x >= head_size)
        ):
            return check_result.fail(
                "x is not being split into two equal halves of size half_head_size."
            )
        return check_result

    def rewrite(self, op, x, cos, sin, **_):
        num_heads = x.shape[1]
        return op.RotaryEmbedding(
            x, cos, sin, interleaved=0, num_heads=num_heads, _domain="ai.onnxruntime._fusion"
        )


class PartialRotaryEmbeddingFusion(pattern.RewriteRuleClassBase):
    def pattern(self, op, x, end1, start2):
        x_part_1 = op.Slice(x, [0], end1, [3], [1])
        x_part_2 = op.Slice(x, start2, [9223372036854775807], [3], [1])
        x_part_1_rope = op.RotaryEmbedding(
            x_part_1,
            _allow_other_inputs=True,
            _allow_other_attributes=True,
            _domain="com.microsoft",
            _outputs=["x_part_1_rope"],
        )
        return op.Concat(x_part_1_rope, x_part_2, axis=-1)

    def check(self, op, x, end1, start2, x_part_1_rope, **_) -> pattern.MatchResult:  # type: ignore[name-defined]
        check_result = pattern.MatchResult()
        end1_value = _ir_utils.get_singleton_value(end1)
        start2_value = _ir_utils.get_singleton_value(start2)
        if not isinstance(end1_value, int) or not isinstance(start2_value, int):
            return check_result.fail(
                "The end1 value of first slice and start2 value of second slice are not integers."
            )
        if end1_value != start2_value:
            return check_result.fail(
                "The end1 value of first slice and start2 value of second slice are not equal."
            )
        rotary_embedding_attributes = x_part_1_rope.producer().attributes
        if "rotary_embedding_dim" in rotary_embedding_attributes:
            return check_result.fail("rotary_embedding_dim attribute already specified.")
        if (
            "interleaved" in rotary_embedding_attributes
            and rotary_embedding_attributes["interleaved"].value != 0
        ):
            return check_result.fail("interleaved is not equal to 0.")
        return check_result

    def rewrite(self, op, x, end1, x_part_1_rope, **_):
        # Create a modified version of the RotaryEmbedding op:
        rotary_embedding_dim = _ir_utils.get_singleton_value(end1)
        original_node = x_part_1_rope.producer()
        inputs = list(original_node.inputs)
        inputs[0] = x
        attrs = dict(original_node.attributes)
        attrs["rotary_embedding_dim"] = rotary_embedding_dim
        return op.RotaryEmbedding(
            *inputs,
            **attrs,
            _domain="com.microsoft",
        )


_rule = RotaryEmbeddingFusion.rule()

_partial_embedding_rule = PartialRotaryEmbeddingFusion.rule()

rotary_embedding_rules = pattern.RewriteRuleSet([_rule])

partial_embedding_rules = pattern.RewriteRuleSet([_partial_embedding_rule])


fuse_rotary_embedding = _fusion_utils.apply_fusion_rules(rotary_embedding_rules)


fuse_partial_rotary_embedding = _fusion_utils.apply_fusion_rules(partial_embedding_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import copy
import dataclasses
import logging
from typing import Dict, Mapping, Optional, Sequence, Set

import onnx
import onnx.defs

import onnxscript
from onnxscript import ir

logger = logging.getLogger(__name__)

_ONNX_DTYPE_TO_ONNX_TENSOR_TYPE_STR: Mapping[int, str] = {
    onnx.TensorProto.FLOAT: "tensor(float)",
    onnx.TensorProto.UINT8: "tensor(uint8)",
    onnx.TensorProto.INT8: "tensor(int8)",
    onnx.TensorProto.UINT16: "tensor(uint16)",
    onnx.TensorProto.INT16: "tensor(int16)",
    onnx.TensorProto.INT32: "tensor(int32)",
    onnx.TensorProto.INT64: "tensor(int64)",
    onnx.TensorProto.STRING: "tensor(string)",
    onnx.TensorProto.BOOL: "tensor(bool)",
    onnx.TensorProto.FLOAT16: "tensor(float16)",
    onnx.TensorProto.DOUBLE: "tensor(double)",
    onnx.TensorProto.COMPLEX64: "tensor(complex64)",
    onnx.TensorProto.COMPLEX128: "tensor(complex128)",
    onnx.TensorProto.UINT32: "tensor(uint32)",
    onnx.TensorProto.UINT64: "tensor(uint64)",
    onnx.TensorProto.BFLOAT16: "tensor(bfloat16)",
}

_ONNX_ATTR_TYPE_TO_ONNX_TENSOR_TYPE_STR: Mapping[int, str] = {
    onnx.AttributeProto.UNDEFINED: "undefined",
    onnx.AttributeProto.FLOAT: "tensor(float)",
    onnx.AttributeProto.INT: "tensor(int64)",
    onnx.AttributeProto.STRING: "tensor(string)",
    onnx.AttributeProto.FLOATS: "tensor(float)",
    onnx.AttributeProto.INTS: "tensor(int64)",
}


class TypeConstraint:
    """Type constraint shared by multiple values."""

    name: str
    type_strs: Set[str]
    values: Set[Value]

    def __init__(self, name: str, type_strs: Set[str]):
        self.name = name
        self.type_strs = type_strs
        self.values = set()

    def merge_type_constraint(self, other: TypeConstraint):
        """Merge two type constraints.

        Typically this process tightens the type constraints.

        For example, consider two values with different existing type constraints:

            input1: TypeConstraint(name="input1", type_strs={"tensor(float)", "tensor(int64)"})
            input2: TypeConstraint(name="input2", type_strs={"tensor(float)", "tensor(int32)"})

        Now consider they are two inputs to `Add` node, where the op schema is:

            inputs: [T, T]
            outputs: [T]
            T: ['tensor(uint8)', ..., 'tensor(float)', 'tensor(int64)', 'tensor(int32)']

        `input1` and `input2` must now bound to the same type constraint, hence the result is
        the intersection of the two existing type constraints with the op schema type constraint:

            input1: TypeConstraint(name="input1.input2", type_strs={"tensor(float)"})
            input2: TypeConstraint(name="input1.input2", type_strs={"tensor(float)"})
        """
        if self is other:
            return
        self.type_strs.intersection_update(other.type_strs)
        self.values = self.values.union(other.values)
        for meta in self.values:
            meta.type_constraint = self
        self.name = f"{self.name}.{other.name}"

    def bind_value(self, value: Value):
        """Bind a value to this type constraint.

        If the value does not have a type constraint, set it to this type constraint.
        Otherwise, merge the two type constraints.
        """
        if value.type_constraint is not None:
            self.merge_type_constraint(value.type_constraint)
        else:
            self.values.add(value)
            value.type_constraint = self

    def __repr__(self):
        value_names = [value.name for value in self.values]
        return f"TypeConstraint(name={self.name}, type_strs={self.type_strs}, values={value_names})"


class Value:
    """Represents a value in the graph. Associated with a TypeConstraint."""

    def __init__(self, name: str):
        self.type_constraint: Optional[TypeConstraint] = None
        self.name: str = name

    def merge_type_constraint(self, other: Value):
        if other.type_constraint is not None:
            other.type_constraint.bind_value(self)
        elif self.type_constraint is not None:
            self.type_constraint.bind_value(other)
        else:
            raise ValueError(
                f"Cannot merge two values without type constraints. {self} and {other}"
            )

    def __repr__(self) -> str:
        return f"Value(name={self.name}, type_constraint={self.type_constraint})"


@dataclasses.dataclass
class OnnxFunctionTypeConstraints:
    input_type_constraints: Dict[str, Optional[TypeConstraint]]
    output_type_constraints: Dict[str, Optional[TypeConstraint]]
    intermediate_type_constraints: Dict[str, Optional[TypeConstraint]]

    def __repr__(self):
        repr_strs = [
            "Type Constraints:",
            "  Inputs: ",
        ]
        repr_strs += [
            f"    {name}: {type_constraint.name}"
            if type_constraint is not None
            else f"    {name}: None"
            for name, type_constraint in self.input_type_constraints.items()
        ]
        repr_strs += [
            "  Outputs: ",
        ]
        repr_strs += [
            f"    {name}: {type_constraint.name}"
            if type_constraint is not None
            else f"    {name}: None"
            for name, type_constraint in self.output_type_constraints.items()
        ]
        repr_strs += [
            "  Type Constraints: ",
        ]
        # Trick to get unique type constraints but maintain the order.
        ordered_unique_type_constraints = dict.fromkeys(self.input_type_constraints.values())
        ordered_unique_type_constraints.update(
            dict.fromkeys(self.output_type_constraints.values())
        )
        repr_strs += [
            f"    {type_constraint.name}: {type_constraint.type_strs}"
            for type_constraint in ordered_unique_type_constraints
            if type_constraint is not None
        ]

        if self.intermediate_type_constraints:
            repr_strs += ["  Intermediate Values: "]
            repr_strs += [
                f"    {name}: {type_constraint.name}"
                if type_constraint is not None
                else f"    {name}: None"
                for name, type_constraint in self.intermediate_type_constraints.items()
            ]

            repr_strs += [
                "  Intermediate Type Constraints: ",
            ]
            ordered_unique_type_constraints = dict.fromkeys(
                self.intermediate_type_constraints.values()
            )
            repr_strs += [
                f"    {type_constraint.name}: {type_constraint.type_strs}"
                for type_constraint in ordered_unique_type_constraints
                if type_constraint is not None
            ]

        return "\n".join(repr_strs)


class TypeConstraintDeducer:
    def __init__(self, onnx_function: onnxscript.OnnxFunction):
        self.onnx_function = onnx_function
        self.values: Dict[str, Value] = {}

    def type_constraints(self, signature_only: bool = True) -> OnnxFunctionTypeConstraints:
        """Retrieve deduced type constraints for the ONNX function."""
        if not self.values:
            raise ValueError("Must call deduce() first")

        input_type_constraints = {
            name: self.values[name].type_constraint for name in self.function_proto.input
        }
        output_type_constraints = {
            name: self.values[name].type_constraint for name in self.function_proto.output
        }
        intermediate_type_constraints = (
            {}
            if signature_only
            else {name: value.type_constraint for name, value in self.values.items()}
        )

        # Rename type constraints to T0, T1, T2, ...
        seen_type_constraints: Set[TypeConstraint] = set()
        for type_constraint in (
            *input_type_constraints.values(),
            *output_type_constraints.values(),
            *intermediate_type_constraints.values(),
        ):
            if type_constraint is not None and type_constraint not in seen_type_constraints:
                type_constraint.name = f"T{len(seen_type_constraints)}"
                seen_type_constraints.add(type_constraint)

        return OnnxFunctionTypeConstraints(
            input_type_constraints, output_type_constraints, intermediate_type_constraints
        )

    def deduce(self, signature_only: bool = True) -> OnnxFunctionTypeConstraints:
        """Deduce type constraints for all values in the graph.

        Args:
            signature_only: If True, return deduce type constraints only for function signature.
                Otherwise, return deduced type constraints for all values in the graph.
        """
        self.values = {}
        self.function_proto = self.onnx_function.to_function_proto()
        self.opset_version = self.function_proto.opset_import[0].version

        op_signature = self.onnx_function.op_signature
        if op_signature is not None:
            for param in op_signature.params:
                if isinstance(param, ir.schemas.Parameter):
                    self.values[param.name] = Value(param.name)

        for node in self.function_proto.node:
            self._process_node(node)

        return self.type_constraints(signature_only)

    def _bind_signature(
        self,
        node: onnx.NodeProto,
        param_names: Sequence[str],
        params: Sequence[ir.schemas.Parameter],
        op_type_constraints: Dict[str, TypeConstraint],
        is_output: bool = False,
    ):
        params_list = list(params)
        # If the last parameter is variadic, duplicate it to match the number of parameters.
        if len(params_list) < len(param_names) and params_list[-1].variadic:
            params_list += [params_list[-1]] * (len(param_names) - len(params_list))
        for name, param in zip(param_names, params_list):
            if is_output:
                if name in self.values:
                    raise ValueError(f"Output {name} already exists.")
                value = Value(name)
                self.values[name] = value
            else:
                if not name:
                    # Skip optional inputs
                    continue
                value = self.values[name]
            if (
                new_type_constraint := op_type_constraints.get(param.type_constraint.name)
            ) is None:
                # parameter is annotated with type string instead of type constraint.
                # Create individual type constraint.
                # Convert allowed_types from Parameter.type_constraint to type_strs
                type_strs = {str(t) for t in param.type_constraint.allowed_types}
                new_type_constraint = TypeConstraint(
                    name=f"T_{name}",
                    type_strs=type_strs,
                )
            # Note: OpSignature Parameter always has is_homogeneous=True for non-variadic params
            # For variadic params, each occurrence gets its own type constraint
            if param.variadic:
                # Parameter is variadic, create a type constraint copy.
                # Meaning the type constraint is not shared amongst them.
                new_type_constraint = TypeConstraint(
                    name=f"{param.name}_{name}",
                    type_strs=new_type_constraint.type_strs,
                )
            prev_value_constraint = copy.deepcopy(value.type_constraint)
            new_type_constraint.bind_value(value)
            if prev_value_constraint is not None and len(
                prev_value_constraint.type_strs
            ) > len(new_type_constraint.type_strs):
                logger.info(
                    "Type constraint is tightened due to binding %s with parameter %s in node %s(%s)",
                    value.name,
                    param.name,
                    node.op_type,
                    node.name,
                )
                logger.info("  %s", prev_value_constraint.type_strs)
                logger.info("->")
                logger.info("  %s", new_type_constraint.type_strs)

    def _perform_extra_type_constraint_tightening(self, node: onnx.NodeProto):
        if node.op_type == "If":
            # Step into subgraph for more type constraint deduction.
            subgraph_attr_names = ("then_branch", "else_branch")
            for attr in node.attribute:
                if attr.name not in subgraph_attr_names:
                    continue
                subgraph = attr.g
                for subgraph_node in subgraph.node:
                    self._process_node(subgraph_node)
                for subgraph_output, node_output_name in zip(subgraph.output, node.output):
                    # Type constraint must agree between "then" and "else" branch.
                    self.values[node_output_name].merge_type_constraint(
                        self.values[subgraph_output.name]
                    )
        elif node.op_type in ("Loop", "Scan"):
            # Step into subgraph for more type constraint deduction.
            raise NotImplementedError("Loop/Scan is not supported yet!")
        elif node.op_type == "Constant":
            # Constant type is static and can be inferred from attribute.
            # Tighten the type constraint.
            tensor_attr = node.attribute[0]
            type_constraint_name = f"Constant_{node.name}"
            if tensor_attr.type == onnx.AttributeProto.TENSOR:
                type_constraint = TypeConstraint(
                    type_constraint_name,
                    {_ONNX_DTYPE_TO_ONNX_TENSOR_TYPE_STR[tensor_attr.t.data_type]},
                )
            else:
                type_constraint = TypeConstraint(
                    type_constraint_name,
                    {_ONNX_ATTR_TYPE_TO_ONNX_TENSOR_TYPE_STR[tensor_attr.type]},
                )
            type_constraint.bind_value(self.values[node.output[0]])
        elif node.op_type == "Cast":
            to_attr = node.attribute[0]
            if not to_attr.ref_attr_name:
                # Cast to a static type.
                # Tighten the type constraint
                type_constraint = TypeConstraint(
                    f"Cast_{node.name}",
                    {_ONNX_DTYPE_TO_ONNX_TENSOR_TYPE_STR[node.attribute[0].i]},
                )
                type_constraint.bind_value(self.values[node.output[0]])

    def _process_node(self, node: onnx.NodeProto):
        if node.domain and node.domain != "onnx":
            raise NotImplementedError("Nested function is not supported yet.")

        if node.op_type in ("Loop", "Scan"):
            # Step into subgraph for more type constraint deduction.
            raise NotImplementedError("Loop/Scan is not supported yet!")

        # Creating new type constraints from op schema
        op_schema = onnx.defs.get_schema(
            node.op_type, max_inclusive_version=self.opset_version, domain=node.domain
        )
        op_signature = ir.schemas.OpSignature.from_op_schema(op_schema)

        op_type_constraints = {
            param.type_constraint.name: TypeConstraint(
                name=f"{param.type_constraint.name}_{node.name}",
                type_strs={str(t) for t in param.type_constraint.allowed_types},
            )
            for param in op_signature.params
            if isinstance(param, ir.schemas.Parameter)
        }
        # Add output type constraints
        for output_param in op_signature.outputs:
            op_type_constraints[output_param.type_constraint.name] = TypeConstraint(
                name=f"{output_param.type_constraint.name}_{node.name}",
                type_strs={str(t) for t in output_param.type_constraint.allowed_types},
            )

        # Binding new type constraints to input values.
        self._bind_signature(node, node.input, op_signature.inputs, op_type_constraints)
        # Creating new values for outputs, and bind with type constraints.
        self._bind_signature(
            node,
            node.output,
            op_signature.outputs,
            op_type_constraints,
            is_output=True,
        )

        # Postprocess a few special cases
        self._perform_extra_type_constraint_tightening(node)


def deduce_type_constraints(
    onnx_function: onnxscript.OnnxFunction, signature_only: bool = True
) -> OnnxFunctionTypeConstraints:
    """Deduce type constraints for an ONNX function.

    * Expects Tensors to be pre-annotated as `TensorType`.
    * Expects Attributes to be pre-annotated correctly.
    * Produces the least strict type constraints allowed by deducing from OpSchema in FunctionProto.
    * Produces explanation for type constraint tightening.
    * `TracedOnnxFunction` is not planned.

    Args:
        onnx_function: ONNX function to generate type constraints for.
        signature_only: Whether to only generate type constraints for the function signature.

    Returns:
        Type constraints for the ONNX function.
    """
    logger.info("Deducing type constraints for %s", onnx_function.name)
    return TypeConstraintDeducer(onnx_function).deduce(signature_only=signature_only)



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: N801, D417
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Mapping, Optional, Sequence, Tuple, TypeVar, Union

from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_types import DOUBLE, FLOAT, INT32, INT64, STRING
from onnxscript.values import Op, Opset


class Opset_ai_onnx_ml1(Opset):
    def __new__(cls):
        return Opset.__new__(cls, "ai.onnx.ml", 1)

    T_ArrayFeatureExtractor = TypeVar(
        "T_ArrayFeatureExtractor", DOUBLE, FLOAT, INT32, INT64, STRING
    )

    def ArrayFeatureExtractor(
        self, X: T_ArrayFeatureExtractor, Y: INT64
    ) -> T_ArrayFeatureExtractor:
        r"""[🌐 ai.onnx.ml::ArrayFeatureExtractor(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_ArrayFeatureExtractor.html#arrayfeatureextractor-1 "Online Documentation")


            Select elements of the input tensor based on the indices passed.

            The indices are applied to the last axes of the tensor.


        Args:
            X: Data to be selected

            Y: The indices, based on 0 as the first index of any dimension.
        """

        schema = get_schema("ArrayFeatureExtractor", 1, "ai.onnx.ml")
        op = Op(self, "ArrayFeatureExtractor", schema)
        return op(*self._prepare_inputs(schema, X, Y))

    T_Binarizer = TypeVar("T_Binarizer", DOUBLE, FLOAT, INT32, INT64)

    def Binarizer(self, X: T_Binarizer, *, threshold: float = 0.0) -> T_Binarizer:
        r"""[🌐 ai.onnx.ml::Binarizer(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_Binarizer.html#binarizer-1 "Online Documentation")


            Maps the values of the input tensor to either 0 or 1, element-wise, based on the outcome of a comparison against a threshold value.


        Args:
            X: Data to be binarized

            threshold: Values greater than this are mapped to 1, others to 0.
        """

        schema = get_schema("Binarizer", 1, "ai.onnx.ml")
        op = Op(self, "Binarizer", schema)
        return op(*self._prepare_inputs(schema, X), threshold=threshold)

    T1_CastMap = TypeVar("T1_CastMap", Mapping[int, FLOAT], Mapping[int, STRING])

    T2_CastMap: TypeAlias = Union[FLOAT, INT64, STRING]

    def CastMap(
        self,
        X: T1_CastMap,
        *,
        cast_to: str = "TO_FLOAT",
        map_form: str = "DENSE",
        max_map: int = 1,
    ) -> T2_CastMap:
        r"""[🌐 ai.onnx.ml::CastMap(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_CastMap.html#castmap-1 "Online Documentation")


            Converts a map to a tensor.
        The map key must be an int64 and the values will be ordered
            in ascending order based on this key.
        The operator supports dense packing or sparse packing.
            If using sparse packing, the key cannot exceed the max_map-1 value.


        Args:
            X: The input map that is to be cast to a tensor

            cast_to: A string indicating the desired element type of the output tensor,
                one of 'TO_FLOAT', 'TO_STRING', 'TO_INT64'.

            map_form: Indicates whether to only output as many values as are in the
                input (dense), or position the input based on using the key of the map
                as the index of the output (sparse).<br>One of 'DENSE', 'SPARSE'.

            max_map: If the value of map_form is 'SPARSE,' this attribute indicates the
                total length of the output tensor.
        """

        schema = get_schema("CastMap", 1, "ai.onnx.ml")
        op = Op(self, "CastMap", schema)
        return op(
            *self._prepare_inputs(schema, X),
            cast_to=cast_to,
            map_form=map_form,
            max_map=max_map,
        )

    T1_CategoryMapper = TypeVar("T1_CategoryMapper", INT64, STRING)

    T2_CategoryMapper: TypeAlias = Union[INT64, STRING]

    def CategoryMapper(
        self,
        X: T1_CategoryMapper,
        *,
        cats_int64s: Optional[Sequence[int]] = None,
        cats_strings: Optional[Sequence[str]] = None,
        default_int64: int = -1,
        default_string: str = "_Unused",
    ) -> T2_CategoryMapper:
        r"""[🌐 ai.onnx.ml::CategoryMapper(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_CategoryMapper.html#categorymapper-1 "Online Documentation")


            Converts strings to integers and vice versa.

            Two sequences of equal length are used to map between integers and strings,
            with strings and integers at the same index detailing the mapping.

            Each operator converts either integers to strings or strings to integers, depending
            on which default value attribute is provided. Only one default value attribute
            should be defined.

            If the string default value is set, it will convert integers to strings.
            If the int default value is set, it will convert strings to integers.


        Args:
            X: Input data

            cats_int64s: The integers of the map. This sequence must be the same length
                as the 'cats_strings' sequence.

            cats_strings: The strings of the map. This sequence must be the same length
                as the 'cats_int64s' sequence

            default_int64: An integer to use when an input string value is not found in
                the map.<br>One and only one of the 'default_*' attributes must be
                defined.

            default_string: A string to use when an input integer value is not found in
                the map.<br>One and only one of the 'default_*' attributes must be
                defined.
        """

        schema = get_schema("CategoryMapper", 1, "ai.onnx.ml")
        op = Op(self, "CategoryMapper", schema)
        return op(
            *self._prepare_inputs(schema, X),
            cats_int64s=cats_int64s,
            cats_strings=cats_strings,
            default_int64=default_int64,
            default_string=default_string,
        )

    T1_DictVectorizer = TypeVar(
        "T1_DictVectorizer",
        Mapping[int, DOUBLE],
        Mapping[int, FLOAT],
        Mapping[int, STRING],
        Mapping[str, DOUBLE],
        Mapping[str, FLOAT],
        Mapping[str, INT64],
    )

    T2_DictVectorizer: TypeAlias = Union[DOUBLE, FLOAT, INT64, STRING]

    def DictVectorizer(
        self,
        X: T1_DictVectorizer,
        *,
        int64_vocabulary: Optional[Sequence[int]] = None,
        string_vocabulary: Optional[Sequence[str]] = None,
    ) -> T2_DictVectorizer:
        r"""[🌐 ai.onnx.ml::DictVectorizer(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_DictVectorizer.html#dictvectorizer-1 "Online Documentation")


            Uses an index mapping to convert a dictionary to an array.

            Given a dictionary, each key is looked up in the vocabulary attribute corresponding to
            the key type. The index into the vocabulary array at which the key is found is then
            used to index the output 1-D tensor 'Y' and insert into it the value found in the dictionary 'X'.

            The key type of the input map must correspond to the element type of the defined vocabulary attribute.
            Therefore, the output array will be equal in length to the index mapping vector parameter.
            All keys in the input dictionary must be present in the index mapping vector.
            For each item in the input dictionary, insert its value in the output array.
            Any keys not present in the input dictionary, will be zero in the output array.

            For example: if the ``string_vocabulary`` parameter is set to ``["a", "c", "b", "z"]``,
            then an input of ``{"a": 4, "c": 8}`` will produce an output of ``[4, 8, 0, 0]``.


        Args:
            X: A dictionary.

            int64_vocabulary: An integer vocabulary array.<br>One and only one of the
                vocabularies must be defined.

            string_vocabulary: A string vocabulary array.<br>One and only one of the
                vocabularies must be defined.
        """

        schema = get_schema("DictVectorizer", 1, "ai.onnx.ml")
        op = Op(self, "DictVectorizer", schema)
        return op(
            *self._prepare_inputs(schema, X),
            int64_vocabulary=int64_vocabulary,
            string_vocabulary=string_vocabulary,
        )

    T1_FeatureVectorizer = TypeVar("T1_FeatureVectorizer", DOUBLE, FLOAT, INT32, INT64)

    def FeatureVectorizer(
        self, *X: T1_FeatureVectorizer, inputdimensions: Optional[Sequence[int]] = None
    ) -> FLOAT:
        r"""[🌐 ai.onnx.ml::FeatureVectorizer(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_FeatureVectorizer.html#featurevectorizer-1 "Online Documentation")


            Concatenates input tensors into one continuous output.

            All input shapes are 2-D and are concatenated along the second dimension. 1-D tensors are treated as [1,C].
            Inputs are copied to the output maintaining the order of the input arguments.

            All inputs must be integers or floats, while the output will be all floating point values.


        Args:
            X: (variadic) An ordered collection of tensors, all with the same element
                type.

            inputdimensions: The size of each input in the input list
        """

        schema = get_schema("FeatureVectorizer", 1, "ai.onnx.ml")
        op = Op(self, "FeatureVectorizer", schema)
        return op(*self._prepare_inputs(schema, *X), inputdimensions=inputdimensions)

    T_Imputer = TypeVar("T_Imputer", DOUBLE, FLOAT, INT32, INT64)

    def Imputer(
        self,
        X: T_Imputer,
        *,
        imputed_value_floats: Optional[Sequence[float]] = None,
        imputed_value_int64s: Optional[Sequence[int]] = None,
        replaced_value_float: float = 0.0,
        replaced_value_int64: int = 0,
    ) -> T_Imputer:
        r"""[🌐 ai.onnx.ml::Imputer(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_Imputer.html#imputer-1 "Online Documentation")


            Replaces inputs that equal one value with another, leaving all other elements alone.

            This operator is typically used to replace missing values in situations where they have a canonical
            representation, such as -1, 0, NaN, or some extreme value.

            One and only one of imputed_value_floats or imputed_value_int64s should be defined -- floats if the input tensor
            holds floats, integers if the input tensor holds integers. The imputed values must all fit within the
            width of the tensor element type. One and only one of the replaced_value_float or replaced_value_int64 should be defined,
            which one depends on whether floats or integers are being processed.

            The imputed_value attribute length can be 1 element, or it can have one element per input feature.
        In other words, if the input tensor has the shape [*,F], then the length of the attribute array may be 1 or F. If it is 1, then it is broadcast along the last dimension and applied to each feature.


        Args:
            X: Data to be processed.

            imputed_value_floats: Value(s) to change to

            imputed_value_int64s: Value(s) to change to.

            replaced_value_float: A value that needs replacing.

            replaced_value_int64: A value that needs replacing.
        """

        schema = get_schema("Imputer", 1, "ai.onnx.ml")
        op = Op(self, "Imputer", schema)
        return op(
            *self._prepare_inputs(schema, X),
            imputed_value_floats=imputed_value_floats,
            imputed_value_int64s=imputed_value_int64s,
            replaced_value_float=replaced_value_float,
            replaced_value_int64=replaced_value_int64,
        )

    T1_LabelEncoder = TypeVar("T1_LabelEncoder", INT64, STRING)

    T2_LabelEncoder: TypeAlias = Union[INT64, STRING]

    def LabelEncoder(
        self,
        X: T1_LabelEncoder,
        *,
        classes_strings: Optional[Sequence[str]] = None,
        default_int64: int = -1,
        default_string: str = "_Unused",
    ) -> T2_LabelEncoder:
        r"""[🌐 ai.onnx.ml::LabelEncoder(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_LabelEncoder.html#labelencoder-1 "Online Documentation")


            Converts strings to integers and vice versa.

            If the string default value is set, it will convert integers to strings.
            If the int default value is set, it will convert strings to integers.

            Each operator converts either integers to strings or strings to integers, depending
            on which default value attribute is provided. Only one default value attribute
            should be defined.

            When converting from integers to strings, the string is fetched from the
            'classes_strings' list, by simple indexing.

            When converting from strings to integers, the string is looked up in the list
            and the index at which it is found is used as the converted value.


        Args:
            X: Input data.

            classes_strings: A list of labels.

            default_int64: An integer to use when an input string value is not found in
                the map.<br>One and only one of the 'default_*' attributes must be
                defined.

            default_string: A string to use when an input integer value is not found in
                the map.<br>One and only one of the 'default_*' attributes must be
                defined.
        """

        schema = get_schema("LabelEncoder", 1, "ai.onnx.ml")
        op = Op(self, "LabelEncoder", schema)
        return op(
            *self._prepare_inputs(schema, X),
            classes_strings=classes_strings,
            default_int64=default_int64,
            default_string=default_string,
        )

    T1_LinearClassifier = TypeVar("T1_LinearClassifier", DOUBLE, FLOAT, INT32, INT64)

    T2_LinearClassifier: TypeAlias = Union[INT64, STRING]

    def LinearClassifier(
        self,
        X: T1_LinearClassifier,
        *,
        classlabels_ints: Optional[Sequence[int]] = None,
        classlabels_strings: Optional[Sequence[str]] = None,
        coefficients: Sequence[float],
        intercepts: Optional[Sequence[float]] = None,
        multi_class: int = 0,
        post_transform: str = "NONE",
    ) -> Tuple[T2_LinearClassifier, FLOAT]:
        r"""[🌐 ai.onnx.ml::LinearClassifier(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_LinearClassifier.html#linearclassifier-1 "Online Documentation")


            Linear classifier


        Args:
            X: Data to be classified.

            classlabels_ints: Class labels when using integer labels. One and only one
                'classlabels' attribute must be defined.

            classlabels_strings: Class labels when using string labels. One and only one
                'classlabels' attribute must be defined.

            coefficients: A collection of weights of the model(s).

            intercepts: A collection of intercepts.

            multi_class: Indicates whether to do OvR or multinomial (0=OvR is the
                default).

            post_transform: Indicates the transform to apply to the scores
                vector.<br>One of 'NONE,' 'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or
                'PROBIT'
        """

        schema = get_schema("LinearClassifier", 1, "ai.onnx.ml")
        op = Op(self, "LinearClassifier", schema)
        return op(
            *self._prepare_inputs(schema, X),
            classlabels_ints=classlabels_ints,
            classlabels_strings=classlabels_strings,
            coefficients=coefficients,
            intercepts=intercepts,
            multi_class=multi_class,
            post_transform=post_transform,
        )

    T_LinearRegressor = TypeVar("T_LinearRegressor", DOUBLE, FLOAT, INT32, INT64)

    def LinearRegressor(
        self,
        X: T_LinearRegressor,
        *,
        coefficients: Optional[Sequence[float]] = None,
        intercepts: Optional[Sequence[float]] = None,
        post_transform: str = "NONE",
        targets: int = 1,
    ) -> FLOAT:
        r"""[🌐 ai.onnx.ml::LinearRegressor(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_LinearRegressor.html#linearregressor-1 "Online Documentation")


            Generalized linear regression evaluation.

            If targets is set to 1 (default) then univariate regression is performed.

            If targets is set to M then M sets of coefficients must be passed in as a sequence
            and M results will be output for each input n in N.

            The coefficients array is of length n, and the coefficients for each target are contiguous.
            Intercepts are optional but if provided must match the number of targets.


        Args:
            X: Data to be regressed.

            coefficients: Weights of the model(s).

            intercepts: Weights of the intercepts, if used.

            post_transform: Indicates the transform to apply to the regression output
                vector.<br>One of 'NONE,' 'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or
                'PROBIT'

            targets: The total number of regression targets, 1 if not defined.
        """

        schema = get_schema("LinearRegressor", 1, "ai.onnx.ml")
        op = Op(self, "LinearRegressor", schema)
        return op(
            *self._prepare_inputs(schema, X),
            coefficients=coefficients,
            intercepts=intercepts,
            post_transform=post_transform,
            targets=targets,
        )

    T_Normalizer = TypeVar("T_Normalizer", DOUBLE, FLOAT, INT32, INT64)

    def Normalizer(self, X: T_Normalizer, *, norm: str = "MAX") -> FLOAT:
        r"""[🌐 ai.onnx.ml::Normalizer(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_Normalizer.html#normalizer-1 "Online Documentation")


            Normalize the input.  There are three normalization modes, which have the corresponding formulas,
            defined using element-wise infix operators '/' and '^' and tensor-wide functions 'max' and 'sum':



            Max: Y = X / max(X)

            L1:  Y = X / sum(X)

            L2:  Y = sqrt(X^2 / sum(X^2)}

            In all modes, if the divisor is zero, Y == X.


            For batches, that is, [N,C] tensors, normalization is done along the C axis. In other words, each row
            of the batch is normalized independently.


        Args:
            X: Data to be encoded, a tensor of shape [N,C] or [C]

            norm: One of 'MAX,' 'L1,' 'L2'
        """

        schema = get_schema("Normalizer", 1, "ai.onnx.ml")
        op = Op(self, "Normalizer", schema)
        return op(*self._prepare_inputs(schema, X), norm=norm)

    T_OneHotEncoder = TypeVar("T_OneHotEncoder", DOUBLE, FLOAT, INT32, INT64, STRING)

    def OneHotEncoder(
        self,
        X: T_OneHotEncoder,
        *,
        cats_int64s: Optional[Sequence[int]] = None,
        cats_strings: Optional[Sequence[str]] = None,
        zeros: int = 1,
    ) -> FLOAT:
        r"""[🌐 ai.onnx.ml::OneHotEncoder(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_OneHotEncoder.html#onehotencoder-1 "Online Documentation")


            Replace each input element with an array of ones and zeros, where a single
            one is placed at the index of the category that was passed in. The total category count
            will determine the size of the extra dimension of the output array Y.

            For example, if we pass a tensor with a single value of 4, and a category count of 8,
            the output will be a tensor with ``[0,0,0,0,1,0,0,0]``.

            This operator assumes every input feature is from the same set of categories.

            If the input is a tensor of float, int32, or double, the data will be cast
            to integers and the cats_int64s category list will be used for the lookups.


        Args:
            X: Data to be encoded.

            cats_int64s: List of categories, ints.<br>One and only one of the 'cats_*'
                attributes must be defined.

            cats_strings: List of categories, strings.<br>One and only one of the
                'cats_*' attributes must be defined.

            zeros: If true and category is not present, will return all zeros; if false
                and a category if not found, the operator will fail.
        """

        schema = get_schema("OneHotEncoder", 1, "ai.onnx.ml")
        op = Op(self, "OneHotEncoder", schema)
        return op(
            *self._prepare_inputs(schema, X),
            cats_int64s=cats_int64s,
            cats_strings=cats_strings,
            zeros=zeros,
        )

    T1_SVMClassifier = TypeVar("T1_SVMClassifier", DOUBLE, FLOAT, INT32, INT64)

    T2_SVMClassifier: TypeAlias = Union[INT64, STRING]

    def SVMClassifier(
        self,
        X: T1_SVMClassifier,
        *,
        classlabels_ints: Optional[Sequence[int]] = None,
        classlabels_strings: Optional[Sequence[str]] = None,
        coefficients: Optional[Sequence[float]] = None,
        kernel_params: Optional[Sequence[float]] = None,
        kernel_type: str = "LINEAR",
        post_transform: str = "NONE",
        prob_a: Optional[Sequence[float]] = None,
        prob_b: Optional[Sequence[float]] = None,
        rho: Optional[Sequence[float]] = None,
        support_vectors: Optional[Sequence[float]] = None,
        vectors_per_class: Optional[Sequence[int]] = None,
    ) -> Tuple[T2_SVMClassifier, FLOAT]:
        r"""[🌐 ai.onnx.ml::SVMClassifier(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_SVMClassifier.html#svmclassifier-1 "Online Documentation")


            Support Vector Machine classifier


        Args:
            X: Data to be classified.

            classlabels_ints: Class labels if using integer labels.<br>One and only one
                of the 'classlabels_*' attributes must be defined.

            classlabels_strings: Class labels if using string labels.<br>One and only
                one of the 'classlabels_*' attributes must be defined.

            kernel_params: List of 3 elements containing gamma, coef0, and degree, in
                that order. Zero if unused for the kernel.

            kernel_type: The kernel type, one of 'LINEAR,' 'POLY,' 'RBF,' 'SIGMOID'.

            post_transform: Indicates the transform to apply to the score. <br>One of
                'NONE,' 'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT'

            prob_a: First set of probability coefficients.

            prob_b: Second set of probability coefficients. This array must be same size
                as prob_a.<br>If these are provided then output Z are probability
                estimates, otherwise they are raw scores.
        """

        schema = get_schema("SVMClassifier", 1, "ai.onnx.ml")
        op = Op(self, "SVMClassifier", schema)
        return op(
            *self._prepare_inputs(schema, X),
            classlabels_ints=classlabels_ints,
            classlabels_strings=classlabels_strings,
            coefficients=coefficients,
            kernel_params=kernel_params,
            kernel_type=kernel_type,
            post_transform=post_transform,
            prob_a=prob_a,
            prob_b=prob_b,
            rho=rho,
            support_vectors=support_vectors,
            vectors_per_class=vectors_per_class,
        )

    T_SVMRegressor = TypeVar("T_SVMRegressor", DOUBLE, FLOAT, INT32, INT64)

    def SVMRegressor(
        self,
        X: T_SVMRegressor,
        *,
        coefficients: Optional[Sequence[float]] = None,
        kernel_params: Optional[Sequence[float]] = None,
        kernel_type: str = "LINEAR",
        n_supports: int = 0,
        one_class: int = 0,
        post_transform: str = "NONE",
        rho: Optional[Sequence[float]] = None,
        support_vectors: Optional[Sequence[float]] = None,
    ) -> FLOAT:
        r"""[🌐 ai.onnx.ml::SVMRegressor(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_SVMRegressor.html#svmregressor-1 "Online Documentation")


            Support Vector Machine regression prediction and one-class SVM anomaly detection.


        Args:
            X: Data to be regressed.

            coefficients: Support vector coefficients.

            kernel_params: List of 3 elements containing gamma, coef0, and degree, in
                that order. Zero if unused for the kernel.

            kernel_type: The kernel type, one of 'LINEAR,' 'POLY,' 'RBF,' 'SIGMOID'.

            n_supports: The number of support vectors.

            one_class: Flag indicating whether the regression is a one-class SVM or not.

            post_transform: Indicates the transform to apply to the score. <br>One of
                'NONE,' 'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT.'

            support_vectors: Chosen support vectors
        """

        schema = get_schema("SVMRegressor", 1, "ai.onnx.ml")
        op = Op(self, "SVMRegressor", schema)
        return op(
            *self._prepare_inputs(schema, X),
            coefficients=coefficients,
            kernel_params=kernel_params,
            kernel_type=kernel_type,
            n_supports=n_supports,
            one_class=one_class,
            post_transform=post_transform,
            rho=rho,
            support_vectors=support_vectors,
        )

    T_Scaler = TypeVar("T_Scaler", DOUBLE, FLOAT, INT32, INT64)

    def Scaler(
        self,
        X: T_Scaler,
        *,
        offset: Optional[Sequence[float]] = None,
        scale: Optional[Sequence[float]] = None,
    ) -> FLOAT:
        r"""[🌐 ai.onnx.ml::Scaler(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_Scaler.html#scaler-1 "Online Documentation")


            Rescale input data, for example to standardize features by removing the mean and scaling to unit variance.


        Args:
            X: Data to be scaled.

            offset: First, offset by this.<br>Can be length of features in an [N,F]
                tensor or length 1, in which case it applies to all features, regardless
                of dimension count.

            scale: Second, multiply by this.<br>Can be length of features in an [N,F]
                tensor or length 1, in which case it applies to all features, regardless
                of dimension count.<br>Must be same length as 'offset'
        """

        schema = get_schema("Scaler", 1, "ai.onnx.ml")
        op = Op(self, "Scaler", schema)
        return op(*self._prepare_inputs(schema, X), offset=offset, scale=scale)

    T1_TreeEnsembleClassifier = TypeVar(
        "T1_TreeEnsembleClassifier", DOUBLE, FLOAT, INT32, INT64
    )

    T2_TreeEnsembleClassifier: TypeAlias = Union[INT64, STRING]

    def TreeEnsembleClassifier(
        self,
        X: T1_TreeEnsembleClassifier,
        *,
        base_values: Optional[Sequence[float]] = None,
        class_ids: Optional[Sequence[int]] = None,
        class_nodeids: Optional[Sequence[int]] = None,
        class_treeids: Optional[Sequence[int]] = None,
        class_weights: Optional[Sequence[float]] = None,
        classlabels_int64s: Optional[Sequence[int]] = None,
        classlabels_strings: Optional[Sequence[str]] = None,
        nodes_falsenodeids: Optional[Sequence[int]] = None,
        nodes_featureids: Optional[Sequence[int]] = None,
        nodes_hitrates: Optional[Sequence[float]] = None,
        nodes_missing_value_tracks_true: Optional[Sequence[int]] = None,
        nodes_modes: Optional[Sequence[str]] = None,
        nodes_nodeids: Optional[Sequence[int]] = None,
        nodes_treeids: Optional[Sequence[int]] = None,
        nodes_truenodeids: Optional[Sequence[int]] = None,
        nodes_values: Optional[Sequence[float]] = None,
        post_transform: str = "NONE",
    ) -> Tuple[T2_TreeEnsembleClassifier, FLOAT]:
        r"""[🌐 ai.onnx.ml::TreeEnsembleClassifier(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_TreeEnsembleClassifier.html#treeensembleclassifier-1 "Online Documentation")


            Tree Ensemble classifier.  Returns the top class for each of N inputs.

            The attributes named 'nodes_X' form a sequence of tuples, associated by
            index into the sequences, which must all be of equal length. These tuples
            define the nodes.

            Similarly, all fields prefixed with 'class_' are tuples of votes at the leaves.
            A leaf may have multiple votes, where each vote is weighted by
            the associated class_weights index.

            One and only one of classlabels_strings or classlabels_int64s
            will be defined. The class_ids are indices into this list.


        Args:
            X: Input of shape [N,F]

            base_values: Base values for classification, added to final class score; the
                size must be the same as the classes or can be left unassigned (assumed
                0)

            class_ids: The index of the class list that each weight is for.

            class_nodeids: node id that this weight is for.

            class_treeids: The id of the tree that this node is in.

            class_weights: The weight for the class in class_id.

            classlabels_int64s: Class labels if using integer labels.<br>One and only
                one of the 'classlabels_*' attributes must be defined.

            classlabels_strings: Class labels if using string labels.<br>One and only
                one of the 'classlabels_*' attributes must be defined.

            nodes_falsenodeids: Child node if expression is false.

            nodes_featureids: Feature id for each node.

            nodes_hitrates: Popularity of each node, used for performance and may be
                omitted.

            nodes_missing_value_tracks_true: For each node, define what to do in the
                presence of a missing value: if a value is missing (NaN), use the 'true'
                or 'false' branch based on the value in this array.<br>This attribute
                may be left undefined, and the default value is false (0) for all nodes.

            nodes_modes: The node kind, that is, the comparison to make at the node.
                There is no comparison to make at a leaf node.<br>One of 'BRANCH_LEQ',
                'BRANCH_LT', 'BRANCH_GTE', 'BRANCH_GT', 'BRANCH_EQ', 'BRANCH_NEQ',
                'LEAF'

            nodes_nodeids: Node id for each node. Ids may restart at zero for each tree,
                but it not required to.

            nodes_treeids: Tree id for each node.

            nodes_truenodeids: Child node if expression is true.

            nodes_values: Thresholds to do the splitting on for each node.

            post_transform: Indicates the transform to apply to the score. <br> One of
                'NONE,' 'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT.'
        """

        schema = get_schema("TreeEnsembleClassifier", 1, "ai.onnx.ml")
        op = Op(self, "TreeEnsembleClassifier", schema)
        return op(
            *self._prepare_inputs(schema, X),
            base_values=base_values,
            class_ids=class_ids,
            class_nodeids=class_nodeids,
            class_treeids=class_treeids,
            class_weights=class_weights,
            classlabels_int64s=classlabels_int64s,
            classlabels_strings=classlabels_strings,
            nodes_falsenodeids=nodes_falsenodeids,
            nodes_featureids=nodes_featureids,
            nodes_hitrates=nodes_hitrates,
            nodes_missing_value_tracks_true=nodes_missing_value_tracks_true,
            nodes_modes=nodes_modes,
            nodes_nodeids=nodes_nodeids,
            nodes_treeids=nodes_treeids,
            nodes_truenodeids=nodes_truenodeids,
            nodes_values=nodes_values,
            post_transform=post_transform,
        )

    T_TreeEnsembleRegressor = TypeVar("T_TreeEnsembleRegressor", DOUBLE, FLOAT, INT32, INT64)

    def TreeEnsembleRegressor(
        self,
        X: T_TreeEnsembleRegressor,
        *,
        aggregate_function: str = "SUM",
        base_values: Optional[Sequence[float]] = None,
        n_targets: Optional[int] = None,
        nodes_falsenodeids: Optional[Sequence[int]] = None,
        nodes_featureids: Optional[Sequence[int]] = None,
        nodes_hitrates: Optional[Sequence[float]] = None,
        nodes_missing_value_tracks_true: Optional[Sequence[int]] = None,
        nodes_modes: Optional[Sequence[str]] = None,
        nodes_nodeids: Optional[Sequence[int]] = None,
        nodes_treeids: Optional[Sequence[int]] = None,
        nodes_truenodeids: Optional[Sequence[int]] = None,
        nodes_values: Optional[Sequence[float]] = None,
        post_transform: str = "NONE",
        target_ids: Optional[Sequence[int]] = None,
        target_nodeids: Optional[Sequence[int]] = None,
        target_treeids: Optional[Sequence[int]] = None,
        target_weights: Optional[Sequence[float]] = None,
    ) -> FLOAT:
        r"""[🌐 ai.onnx.ml::TreeEnsembleRegressor(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_TreeEnsembleRegressor.html#treeensembleregressor-1 "Online Documentation")


            Tree Ensemble regressor.  Returns the regressed values for each input in N.

            All args with nodes_ are fields of a tuple of tree nodes, and
            it is assumed they are the same length, and an index i will decode the
            tuple across these inputs.  Each node id can appear only once
            for each tree id.

            All fields prefixed with target_ are tuples of votes at the leaves.

            A leaf may have multiple votes, where each vote is weighted by
            the associated target_weights index.

            All trees must have their node ids start at 0 and increment by 1.

            Mode enum is BRANCH_LEQ, BRANCH_LT, BRANCH_GTE, BRANCH_GT, BRANCH_EQ, BRANCH_NEQ, LEAF


        Args:
            X: Input of shape [N,F]

            aggregate_function: Defines how to aggregate leaf values within a target.
                <br>One of 'AVERAGE,' 'SUM,' 'MIN,' 'MAX.'

            base_values: Base values for classification, added to final class score; the
                size must be the same as the classes or can be left unassigned (assumed
                0)

            n_targets: The total number of targets.

            nodes_falsenodeids: Child node if expression is false

            nodes_featureids: Feature id for each node.

            nodes_hitrates: Popularity of each node, used for performance and may be
                omitted.

            nodes_missing_value_tracks_true: For each node, define what to do in the
                presence of a NaN: use the 'true' (if the attribute value is 1) or
                'false' (if the attribute value is 0) branch based on the value in this
                array.<br>This attribute may be left undefined and the default value is
                false (0) for all nodes.

            nodes_modes: The node kind, that is, the comparison to make at the node.
                There is no comparison to make at a leaf node.<br>One of 'BRANCH_LEQ',
                'BRANCH_LT', 'BRANCH_GTE', 'BRANCH_GT', 'BRANCH_EQ', 'BRANCH_NEQ',
                'LEAF'

            nodes_nodeids: Node id for each node. Node ids must restart at zero for each
                tree and increase sequentially.

            nodes_treeids: Tree id for each node.

            nodes_truenodeids: Child node if expression is true

            nodes_values: Thresholds to do the splitting on for each node.

            post_transform: Indicates the transform to apply to the score. <br>One of
                'NONE,' 'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT'

            target_ids: The index of the target that each weight is for

            target_nodeids: The node id of each weight

            target_treeids: The id of the tree that each node is in.

            target_weights: The weight for each target
        """

        schema = get_schema("TreeEnsembleRegressor", 1, "ai.onnx.ml")
        op = Op(self, "TreeEnsembleRegressor", schema)
        return op(
            *self._prepare_inputs(schema, X),
            aggregate_function=aggregate_function,
            base_values=base_values,
            n_targets=n_targets,
            nodes_falsenodeids=nodes_falsenodeids,
            nodes_featureids=nodes_featureids,
            nodes_hitrates=nodes_hitrates,
            nodes_missing_value_tracks_true=nodes_missing_value_tracks_true,
            nodes_modes=nodes_modes,
            nodes_nodeids=nodes_nodeids,
            nodes_treeids=nodes_treeids,
            nodes_truenodeids=nodes_truenodeids,
            nodes_values=nodes_values,
            post_transform=post_transform,
            target_ids=target_ids,
            target_nodeids=target_nodeids,
            target_treeids=target_treeids,
            target_weights=target_weights,
        )

    T_ZipMap: TypeAlias = Union[Sequence[Mapping[int, FLOAT]], Sequence[Mapping[str, FLOAT]]]

    def ZipMap(
        self,
        X: FLOAT,
        *,
        classlabels_int64s: Optional[Sequence[int]] = None,
        classlabels_strings: Optional[Sequence[str]] = None,
    ) -> T_ZipMap:
        r"""[🌐 ai.onnx.ml::ZipMap(1)](https://onnx.ai/onnx/operators/onnx_aionnxml_ZipMap.html#zipmap-1 "Online Documentation")


            Creates a map from the input and the attributes.

            The values are provided by the input tensor, while the keys are specified by the attributes.
            Must provide keys in either classlabels_strings or classlabels_int64s (but not both).

            The columns of the tensor correspond one-by-one to the keys specified by the attributes. There must be as many columns as keys.



        Args:
            X: The input values

            classlabels_int64s: The keys when using int keys.<br>One and only one of the
                'classlabels_*' attributes must be defined.

            classlabels_strings: The keys when using string keys.<br>One and only one of
                the 'classlabels_*' attributes must be defined.
        """

        schema = get_schema("ZipMap", 1, "ai.onnx.ml")
        op = Op(self, "ZipMap", schema)
        return op(
            *self._prepare_inputs(schema, X),
            classlabels_int64s=classlabels_int64s,
            classlabels_strings=classlabels_strings,
        )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import logging

from onnxscript import ir
from onnxscript.rewriter._rewrite_rule import RewriteRule, RewriteRuleSet

logger = logging.getLogger(__name__)


def check_if_not_need_reshape(
    context, input_a: ir.Value, input_b: ir.Value, shape_c: ir.Value, **_
) -> bool:
    """Condition to check if we need to replace the pattern.

    If matmul broadcasting is enough, then we don't need the reshapes.

    To validate this, we need to check the following:
    1. Input shapes check: input_a and input_b should be broadcastable
    2. Output shape check: shape_c should be the same as the output shape from the matmul(input_a, input_b)

    If the above are true, then we don't need the reshapes.

    Returns:
        True if we need to replace the pattern, False otherwise.
    """
    del context  # Reserved for future extensions

    input_a_shape = input_a.shape
    input_b_shape = input_b.shape
    shape_c_tensor = shape_c.const_value
    if shape_c_tensor is None:
        logger.info("The value 'shape_c' is not statically known.")
        return False

    if len(shape_c_tensor.shape) != 1:
        logger.info(
            "Unexpected final shape. The shape of 'shape' value is %s",
            shape_c_tensor.shape,
        )
        return False

    # NOTE: When there is a subset match with a pattern. The MatchResult won't have the shape
    # information. So, we need to check if the shape is None and return False.
    if input_a_shape is None or input_b_shape is None:
        logger.info("Shape information is not available for the inputs and outputs.")
        return False
    if any(isinstance(dim, ir.SymbolicDim) for dim in input_a_shape):
        logger.info("Symbolic dimensions are not yet supported.")
        return False
    if any(isinstance(dim, ir.SymbolicDim) for dim in input_b_shape):
        logger.info("Symbolic dimensions are not yet supported.")
        return False
    input_a_shape = input_a_shape.numpy()  # type: ignore[assignment]
    input_b_shape = input_b_shape.numpy()  # type: ignore[assignment]
    shape_c = shape_c_tensor.numpy().tolist()  # type: ignore[assignment]

    a_rank = len(input_a_shape)
    b_rank = len(input_b_shape)

    # 1. Check if input shapes are broadcastable
    # 1.a. If the first input is 1-D, check whether
    # the dim matches the last second dim of the second input.
    mimic_matmul_broadcast_behavior_a = False
    mimic_matmul_broadcast_behavior_b = False
    if a_rank < 2:
        if b_rank < 2:
            logger.info("Optimization of dot product is not supported yet.")
            return False
        if input_a_shape[-1] != input_b_shape[-2]:
            logger.info("Original shape is not MatMul compatible.")
            return False
        else:
            input_a_shape = [1, *input_a_shape]  # type: ignore[assignment]
            a_rank = len(input_a_shape)
            mimic_matmul_broadcast_behavior_a = True
    # 1.b. If the second input is 1-D, check whether
    # the dim matches the last dim of the first input.
    if b_rank < 2:
        if input_b_shape[-1] != input_a_shape[-1]:
            logger.info("Original shape is not MatMul compatible.")
            return False
        else:
            input_b_shape = [*input_b_shape, 1]  # type: ignore[assignment]
            b_rank = len(input_b_shape)
            mimic_matmul_broadcast_behavior_b = True
    # 1.c. If both inputs are at least 2-D, check whether
    # the last dimension of the first input matches the second
    # last dimension of the second input, and shape[:-2] are
    # broadcastable.
    input_a_shape_except_second_last_dim = [*input_a_shape[:-2], *[input_a_shape[-1]]]
    input_b_shape_except_last_dim = input_b_shape[:-1]
    broadcast_matmul_output_shape = [input_a_shape[-2], input_b_shape[-1]]
    for idx, (dim_from_a, dim_from_b) in enumerate(
        zip(
            reversed(input_a_shape_except_second_last_dim),
            reversed(input_b_shape_except_last_dim),
        )
    ):
        if dim_from_a not in {1, dim_from_b}:
            logger.info("Original shape is not broadcastable.")
            return False
        elif idx > 0:
            broadcast_matmul_output_shape = [
                max(dim_from_a, dim_from_b),  # type: ignore[type-var]
                *broadcast_matmul_output_shape,
            ]

    # 2. Check if output shape is the same as the output shape from the matmul(input_a, input_b)
    # Prepend the broadcast_matmul_output_shape with the longer shape of input
    if a_rank > b_rank:
        longer_shape = input_a_shape
        shorter_shape = input_b_shape
    else:
        longer_shape = input_b_shape
        shorter_shape = input_a_shape
    broadcast_matmul_output_shape = [
        *longer_shape[: -len(shorter_shape)],
        *broadcast_matmul_output_shape,
    ]
    if mimic_matmul_broadcast_behavior_b and b_rank == 2 and input_b_shape[-1] == 1:
        # If input_b is expanded to 2-D, then we need to remove the last dimension
        broadcast_matmul_output_shape = broadcast_matmul_output_shape[:-1]
    if mimic_matmul_broadcast_behavior_a and a_rank == 2 and input_a_shape[0] == 1:
        # If input_a is expanded to 2-D, then we need to remove the first dimension
        # of input_a, which would be the -2nd dimension of the output shape.
        broadcast_matmul_output_shape.pop(-2)
    if shape_c != broadcast_matmul_output_shape:
        logger.info(
            "Final output shape is not the same. Expected %s vs actual %s",
            shape_c,
            broadcast_matmul_output_shape,
        )
        return False

    return True


def _two_reshapes_matmul_reshape_pattern(op, input_a, input_b, shape_a, shape_b, shape_c):
    # TODO: Modified from `value_ints` to `value` to match pattern in benchmark models.
    # This implementation misses pattern of Constants with `value_ints` attribute.
    # See more at https://github.com/microsoft/onnx-rewriter/issues/191.
    # A better solution is to improve pattern matching and avoid depending on writing
    # Constants in pattern. See https://github.com/microsoft/onnx-rewriter/issues/192.
    reshape_a = op.Reshape(input_a, shape_a)
    reshape_b = op.Reshape(input_b, shape_b)
    matmul = op.MatMul(reshape_a, reshape_b)
    return op.Reshape(matmul, shape_c)


def _matmul(op, input_a, input_b, **_):
    return op.MatMul(input_a, input_b)


def _one_reshape_matmul_reshape_pattern(op, input_a, input_b, shape_a, shape_c):
    reshape_a = op.Reshape(input_a, shape_a)
    matmul = op.MatMul(reshape_a, input_b)
    return op.Reshape(matmul, shape_c)


# Register the rewrite rules
two_reshapes_matmul_reshape_rule = RewriteRule(
    _two_reshapes_matmul_reshape_pattern,
    _matmul,
    check_if_not_need_reshape,
)
one_reshape_matmul_reshape_rule = RewriteRule(
    _one_reshape_matmul_reshape_pattern,
    _matmul,
    # We can use the same check_if_not_need_reshape function for both the rules,
    # as one_reshape_matmul_reshape_pattern is a subset of _two_reshapes_matmul_reshape_pattern.
    check_if_not_need_reshape,
)

# NOTE: The order of the rules is important. Larger pattern should be checked first.
rules = RewriteRuleSet([two_reshapes_matmul_reshape_rule, one_reshape_matmul_reshape_rule])



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402, D411
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, TypeVar

from onnx.defs import get_schema

from onnxscript.onnx_opset._impl.opset1 import Opset1
from onnxscript.onnx_types import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset2(Opset1):
    def __new__(cls):
        return Opset.__new__(cls, "", 2)

    T_GlobalLpPool = TypeVar("T_GlobalLpPool", DOUBLE, FLOAT, FLOAT16)

    def GlobalLpPool(self, X: T_GlobalLpPool, *, p: int = 2) -> T_GlobalLpPool:
        r"""[🌐 GlobalLpPool(2)](https://onnx.ai/onnx/operators/onnx__GlobalLpPool.html#globallppool-2 "Online Documentation")


         GlobalLpPool consumes an input tensor X and applies lp pool pooling across
         the values in the same channel. This is equivalent to LpPool with kernel size
         equal to the spatial dimension of input tensor.

        Args:
            X: (differentiable) Input data tensor from the previous operator; dimensions
                for image case are (N x C x H x W), where N is the batch size, C is the
                number of channels, and H and W are the height and the width of the
                data. For non image case, the dimensions are in the form of (N x C x D1
                x D2 ... Dn), where N is the batch size.

            p: p value of the Lp norm used to pool over the input data.
        """

        schema = get_schema("GlobalLpPool", 2, "")
        op = Op(self, "GlobalLpPool", schema)
        return op(*self._prepare_inputs(schema, X), p=p)

    T_LpPool = TypeVar("T_LpPool", DOUBLE, FLOAT, FLOAT16)

    def LpPool(
        self,
        X: T_LpPool,
        *,
        auto_pad: str = "NOTSET",
        kernel_shape: Sequence[int],
        p: int = 2,
        pads: Optional[Sequence[int]] = None,
        strides: Optional[Sequence[int]] = None,
    ) -> T_LpPool:
        r"""[🌐 LpPool(2)](https://onnx.ai/onnx/operators/onnx__LpPool.html#lppool-2 "Online Documentation")


         LpPool consumes an input tensor X and applies Lp pooling across
         the tensor according to kernel sizes, stride sizes, and pad lengths.
         Lp pooling consisting of computing the Lp norm on all values of a subset
         of the input tensor according to the kernel size and downsampling the
         data into the output tensor Y for further processing.

        Args:
            X: Input data tensor from the previous operator; dimensions for image case
                are (N x C x H x W), where N is the batch size, C is the number of
                channels, and H and W are the height and the width of the data. For non
                image case, the dimensions are in the form of (N x C x D1 x D2 ... Dn),
                where N is the batch size.

            auto_pad: auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or VALID.
                Where default value is NOTSET, which means explicit padding is used.
                SAME_UPPER or SAME_LOWER mean pad the input so that the output spatial
                size match the input.In case of odd number add the extra padding at the
                end for SAME_UPPER and at the beginning for SAME_LOWER. VALID mean no
                padding.

            kernel_shape: The size of the kernel along each axis.

            p: p value of the Lp norm used to pool over the input data.

            pads: Padding for the beginning and ending along each spatial axis, it can
                take any value greater than or equal to 0. The value represent the
                number of pixels added to the beginning and end part of the
                corresponding axis. `pads` format should be as follow [x1_begin,
                x2_begin...x1_end, x2_end,...], where xi_begin the number of pixels
                added at the beginning of axis `i` and xi_end, the number of pixels
                added at the end of axis `i`. This attribute cannot be used
                simultaneously with auto_pad attribute. If not present, the padding
                defaults to 0 along start and end of each spatial axis.

            strides: Stride along each spatial axis.
        """

        schema = get_schema("LpPool", 2, "")
        op = Op(self, "LpPool", schema)
        return op(
            *self._prepare_inputs(schema, X),
            auto_pad=auto_pad,
            kernel_shape=kernel_shape,
            p=p,
            pads=pads,
            strides=strides,
        )

    T_Pad = TypeVar("T_Pad", DOUBLE, FLOAT, FLOAT16)

    def Pad(
        self,
        data: T_Pad,
        *,
        mode: str = "constant",
        pads: Sequence[int],
        value: float = 0.0,
    ) -> T_Pad:
        r"""[🌐 Pad(2)](https://onnx.ai/onnx/operators/onnx__Pad.html#pad-2 "Online Documentation")


        Given `data` tensor, pads, mode, and value.
        Example:
          Insert 0 pads to the beginning of the second dimension.
          data = [
              [1.0, 1.2],
              [2.3, 3.4],
              [4.5, 5.7],
          ]
          pads = [0, 2, 0, 0]
          output = [
              [
                  [0.0, 0.0, 1.0, 1.2],
                  [0.0, 0.0, 2.3, 3.4],
                  [0.0, 0.0, 4.5, 5.7],
              ],
          ]


        Args:
            data: Input tensor.

            mode: Three modes: constant(default), reflect, edge

            pads: List of integers indicating the number of padding elements to add or
                remove (if negative) at the beginning and end of each axis. For 2D it is
                the number of pixels. `pads` rank should be double of the input's rank.
                `pads` format should be as follow [x1_begin, x2_begin...x1_end,
                x2_end,...], where xi_begin the number of pixels added at the beginning
                of axis `i` and xi_end, the number of pixels added at the end of axis
                `i`.

            value: One float, indicates the value to be filled.
        """

        schema = get_schema("Pad", 2, "")
        op = Op(self, "Pad", schema)
        return op(*self._prepare_inputs(schema, data), mode=mode, pads=pads, value=value)

    T_Split = TypeVar(
        "T_Split",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Split(
        self, input: T_Split, *, axis: int = 0, split: Optional[Sequence[int]] = None
    ) -> T_Split:
        r"""[🌐 Split(2)](https://onnx.ai/onnx/operators/onnx__Split.html#split-2 "Online Documentation")

        Split a tensor into a list of tensors, along the specified
        'axis'. Lengths of the parts can be specified using argument 'split'.
        Otherwise, the tensor is split to equal sized parts.


        Args:
            input: The tensor to split

            axis: Which axis to split on.

            split: length of each output
        """

        schema = get_schema("Split", 2, "")
        op = Op(self, "Split", schema)
        return op(*self._prepare_inputs(schema, input), axis=axis, split=split)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Union

import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, pattern
from onnxscript.rewriter._basics import MatchFailureError

Dim = Union[int, ir.SymbolicDim]


class SDPAImplementation(pattern.RewriteRuleClassBase):
    def pattern(self, op, query, key, value, key_format):
        """Pattern matches any call to SDPA. See sdpa.py for documentation on the SDPA op."""
        return op.SDPA(
            query,
            key,
            value,
            key_format=key_format,
            _allow_other_inputs=True,  # Mask is optional
            _outputs=["sdpa_output"],
            _domain="ai.onnxruntime._fusion",
        )

    def check(self, context, query, key, value, key_format, sdpa_output):
        bindings: dict[str, Dim] = {}
        _fusion_utils.check_shape(bindings, query, ["B", "H", "S", "Dh"])
        _fusion_utils.check_shape(bindings, value, ["B", "H", "Skv", "Dv"])

        if key_format.value == "BHSd":
            _fusion_utils.check_shape(bindings, key, ["B", "H", "Skv", "Dh"])
        elif key_format.value == "BSHd":
            _fusion_utils.check_shape(bindings, key, ["B", "Skv", "H", "Dh"])
        else:
            raise MatchFailureError(
                f"Unexpected key_format value: {key_format.value}", key_format
            )

        self._num_heads = bindings["H"]
        if not isinstance(self._num_heads, int):
            return False
        self._use_mask_broadcast = True  # TODO: optimize to avoid broadcast if not needed
        return isinstance(self._num_heads, int)

    def rewrite(self, op, query, key, value, key_format, sdpa_output):
        sdpa_node = sdpa_output.producer()
        scale = sdpa_node.attributes.get("scale", None)
        to_3d_shape = op.Constant(value_ints=[0, 0, -1])
        to_4d_shape = op.Constant(value_ints=[0, 0, self._num_heads, -1])
        query_3d = op.Reshape(op.Transpose(query, perm=[0, 2, 1, 3]), to_3d_shape)
        value_3d = op.Reshape(op.Transpose(value, perm=[0, 2, 1, 3]), to_3d_shape)

        if key_format.value == "BHSd":
            key_3d = op.Reshape(op.Transpose(key, perm=[0, 2, 1, 3]), to_3d_shape)
        else:  # BSHd
            key_3d = op.Reshape(key, to_3d_shape)

        inputs = [query_3d, key_3d, value_3d]
        if len(sdpa_node.inputs) > 3:
            mask = sdpa_node.inputs[3]

            if self._use_mask_broadcast:
                one = op.Constant(value_ints=[1])
                query_length = op.Shape(query, start=2, end=3)
                shape_11S1 = op.Concat(one, one, query_length, one, axis=0)
                mask = op.Expand(mask, shape_11S1)

            inputs.extend([None, None, mask])

        output = op.MultiHeadAttention(
            *inputs,
            num_heads=self._num_heads,
            scale=scale,
            _domain="com.microsoft",
        )
        output_4d = op.Reshape(output, to_4d_shape)
        output = op.Transpose(output_4d, perm=[0, 2, 1, 3])
        return output


_rules = pattern.RewriteRuleSet([SDPAImplementation.rule()])

replace_sdpa_by_mha = _fusion_utils.apply_fusion_rules(_rules)
